package Urb::Misc::Sun;

use strict;
use warnings;
use 5.010;
use POSIX;

=head1 NAME

Urb::Misc::Sun - Sun visibility for all latitudes, azimuths and horizon elevations

=head1 SYNOPSIS

A lookup table for calculating the number of minutes of direct sunlight for anywhere

=head1 DESCRIPTION

Each occlusion mask covers 16 azimuth directions (starting north).
For each direction 32 evenly spaced elevations are covered from horizon to zenith.
Values are number of minutes of possible sunlight per year in that direction given an occlusion at that elevation.

For any single location you need to add up the 16 values to get the total number of sun-minutes per year.

=head1 USAGE

  use Urb::Misc::Sun;

=head1 METHODS

=over

=item new

Create a new 'Sun' object, given a latitude:

  my $sun = new Urb::Misc::Sun (53.3814);

=cut

sub new
{
    my $class = shift;
    $class = ref $class || $class;
    my $latitude_deg = shift;
    my $self = Urb::Misc::Sun::table ($latitude_deg);
    bless $self, $class;
    return $self;
}

=item Minutes

Get number of minutes of sun per year in particular direction:

  $minutes = $sun->Minutes ($azimuth_radians, $elevation_radians);

=cut

sub Minutes
{
    my ($self, $azimuth_rad, $elevation_rad) = @_;
    my $azimuth = floor ((azimuths() * $azimuth_rad / 6.283185307) + 0.5);
    $azimuth -= azimuths() if $azimuth >= azimuths();
    $azimuth += azimuths() if $azimuth < 0;

    my $elevation_id = elevations() * $elevation_rad / 1.570796327;
    my $elevation_0 = floor ($elevation_id);
    my $elevation_1 = $elevation_0 +1;

    return ($self->[$azimuth]->[$elevation_0]) if $elevation_0 >= elevations() -1;

    my $weight_0 = $elevation_1 - $elevation_id;
    my $weight_1 = $elevation_id - $elevation_0;

    return ($self->[$azimuth]->[$elevation_0] * $weight_0) + ($self->[$azimuth]->[$elevation_1] * $weight_1);
}

=back

=head1 FUNCTIONS

=over

=item table

Retrieve an interpolated table for a given latitude (in degrees):

  $table_sheffield = Urb::Misc::Sun::table (53.3814);

=cut

sub table
{
    my $latitude_deg = shift;
    my $latitude_id = latitudes() * ($latitude_deg +90) / 180;
    my $latitude_0 = int ($latitude_id);
    my $latitude_1 = $latitude_0 +1;

    my $weight_0 = $latitude_1 - $latitude_id;
    my $weight_1 = $latitude_id - $latitude_0;

    $latitude_0 = 0 if $latitude_0 >= latitudes();
    $latitude_1 = 0 if $latitude_1 >= latitudes();

    my $table_0 = $Urb::Misc::Sun::table->[$latitude_0];
    my $table_1 = $Urb::Misc::Sun::table->[$latitude_1];
    my $table_result;

    for my $azimuth (0 .. azimuths() -1)
    {
        for my $elevation (0 .. elevations() -1)
        {
            $table_result->[$azimuth]->[$elevation]
              = ($table_0->[$azimuth]->[$elevation] * $weight_0)
              + ($table_1->[$azimuth]->[$elevation] * $weight_1);
        }
    }
    return $table_result;
}

=item latitudes azimuths elevations

functions to query the size of the lookup table.

=back

=cut

sub latitudes
{
    return scalar @{$Urb::Misc::Sun::table};
}

sub azimuths
{
    return scalar @{$Urb::Misc::Sun::table->[0]};
}

sub elevations
{
    return scalar @{$Urb::Misc::Sun::table->[0]->[0]};
}

=head1 COPYING
 
Free use of this software is granted under the terms of the GNU General
Public License version 2 or any later version.

=cut

# An occlusion mask for 64 different latitudes (starting at the south pole).
# Each occlusion mask covers 16 azimuth directions (starting north).
# For each direction 32 elevations are covered from horizon to zenith.
# Values are number of minutes of possible sunlight per year in that direction given an occlusion at that elevation.

our $table =
[[[16064,14760,13500,12146,10801,9254,7561,5671,2791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16022,14762,13502,12062,10802,9272,7596,5672,2791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16022,14762,13502,12136,10802,9272,7652,5672,2791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16020,14760,13501,12151,10801,9271,7651,5671,2790,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16109,14850,13590,12240,10852,9360,7740,5759,2877,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16111,14851,13592,12242,10802,9362,7742,5762,2881,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16110,14850,13591,12241,10801,9361,7741,5760,2880,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16150,14850,13590,12240,10800,9360,7740,5760,2880,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16113,14763,13448,12155,10715,9275,7655,5675,2794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16110,14805,13410,12151,10711,9271,7651,5671,2791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16112,14853,13413,12153,10713,9273,7652,5671,2790,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16105,14846,13407,12148,10709,9270,7650,5670,2790,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16111,14849,13462,12152,10712,9273,7640,5673,2791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16112,14762,13502,12152,10802,9272,7562,5671,2707,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16113,14763,13503,12153,10803,9273,7563,5584,2750,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16116,14766,13506,12156,10806,9266,7566,5676,2796,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[17119,15817,14536,13254,11883,10544,9010,7414,5487,2614,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16960,15659,14366,13070,11746,10358,8858,7190,5178,1868,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16676,15387,14101,12790,11436,10014,8489,6744,4546,121,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16298,15005,13704,12381,11011,9550,7950,6079,3513,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15963,14674,13364,12018,10613,9105,7404,5344,1720,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15568,14273,12956,11588,10142,8571,6776,4445,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15260,13958,12630,11241,9771,8145,6227,3599,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15094,13805,12443,11037,9543,7894,5917,3033,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15011,13721,12342,10948,9454,7805,5829,2947,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15171,13864,12555,11148,9681,8057,6142,3505,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15489,14196,12863,11510,10058,8482,6678,4353,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15871,14575,13273,11929,10524,9017,7324,5260,1681,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16297,15013,13718,12388,11016,9555,7954,6084,3517,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16673,15376,14083,12773,11425,10011,8480,6735,4545,120,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16953,15667,14380,13073,11748,10352,8859,7193,5172,1866,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17135,15827,14533,13258,11899,10552,9017,7419,5493,2620,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[18120,16821,15522,14240,12966,11630,10295,8793,7200,5304,2440,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17843,16532,15243,13961,12666,11330,9941,8455,6781,4696,739,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17309,16007,14723,13427,12105,10740,9288,7688,5817,3161,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16560,15267,13970,12658,11298,9861,8300,6507,4164,228,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15807,14514,13199,11842,10420,8880,7132,4937,965,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14990,13680,12335,10925,9412,7722,5683,2291,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14346,13018,11642,10172,8567,6703,4240,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13992,12636,11219,9727,8063,6071,3194,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13896,12545,11131,9636,7972,5981,3104,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14254,12922,11549,10082,8477,6613,4151,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14901,13591,12242,10834,9325,7633,5591,2235,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15724,14428,13115,11756,10332,8793,7047,4857,949,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16557,15269,13976,12654,11294,9861,8300,6507,4168,227,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17304,16005,14716,13426,12106,10737,9288,7690,5809,3155,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17859,16550,15258,13975,12677,11345,9951,8463,6783,4698,737,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18117,16819,15525,14240,12964,11633,10290,8795,7208,5306,2444,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[19183,17796,16496,15206,13935,12667,11347,10022,8559,6990,5122,2270,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18739,17396,16093,14810,13532,12247,10922,9534,8037,6346,4193,283,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17956,16647,15358,14072,12772,11443,10057,8565,6897,4823,1301,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16856,15565,14274,12966,11619,10203,8679,6959,4787,1238,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15674,14381,13055,11684,10239,8668,6865,4499,779,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14432,13105,11727,10259,8656,6794,4312,413,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13396,12021,10567,8983,7166,4837,531,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12785,11377,9875,8207,6208,3338,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12699,11293,9794,8123,6119,3251,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13293,11922,10464,8881,7070,4740,517,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14337,13009,11632,10162,8558,6702,4219,399,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15611,14309,12981,11614,10167,8593,6785,4425,772,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16849,15559,14268,12962,11618,10200,8679,6950,4791,1235,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17955,16640,15353,14069,12771,11440,10055,8566,6889,4820,1300,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18741,17396,16093,14814,13535,12243,10918,9533,8038,6348,4192,282,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19192,17807,16510,15217,13944,12678,11362,10030,8570,6999,5125,2275,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[20251,18810,17425,16133,14850,13590,12339,11042,9739,8299,6764,4930,2093,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19672,18283,16953,15656,14387,13120,11841,10522,9139,7650,5941,3691,77,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18618,17284,15987,14706,13422,12121,10780,9368,7832,6058,3622,378,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17195,15903,14612,13310,11975,10583,9094,7427,5408,2236,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15587,14284,12956,11570,10101,8494,6620,3925,737,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13871,12516,11088,9545,7816,5682,2157,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12382,10945,9378,7607,5385,1423,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11522,10018,8346,6342,3487,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11430,9924,8256,6253,3397,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12287,10850,9283,7510,5291,1379,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13772,12410,10992,9449,7715,5586,2104,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15513,14214,12882,11503,10035,8426,6548,3870,738,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17212,15912,14617,13314,11974,10583,9093,7424,5407,2234,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18626,17303,16007,14728,13445,12142,10798,9383,7845,6067,3626,386,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19660,18262,16933,15642,14367,13103,11825,10508,9127,7635,5932,3688,76,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20250,18810,17427,16136,14856,13593,12343,11048,9743,8307,6773,4934,2102,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[21347,19812,18377,17005,15722,14454,13211,11975,10700,9418,8019,6514,4723,1907,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20645,19179,17802,16483,15201,13939,12685,11420,10114,8742,7255,5540,3155,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19343,17981,16663,15380,14106,12827,11521,10168,8725,7132,5204,2233,17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17573,16274,14990,13686,12360,10984,9522,7919,6023,3183,636,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15532,14225,12880,11482,9994,8348,6393,3427,768,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13302,11908,10422,8781,6858,4043,905,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11324,9783,8043,5917,2294,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10128,8456,6458,3619,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10036,8365,6364,3529,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11227,9683,7949,5822,2241,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13202,11810,10325,8684,6765,3962,881,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15456,14148,12810,11412,9922,8275,6314,3380,769,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17579,16281,14993,13691,12369,10995,9534,7928,6022,3182,638,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19339,17982,16669,15379,14103,12826,11517,10163,8730,7134,5215,2236,18,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20641,19172,17795,16475,15195,13932,12677,11409,10104,8740,7250,5537,3153,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21357,19820,18385,17017,15734,14467,13221,11986,10712,9422,8026,6521,4730,1916,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[22638,20876,19344,17920,16561,15287,14034,12807,11592,10348,9096,7726,6265,4521,1710,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21684,20111,18673,17306,16003,14736,13493,12252,10998,9706,8347,6874,5147,2551,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20080,18682,17349,16054,14782,13510,12231,10921,9552,8084,6417,4242,1370,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17993,16697,15404,14107,12787,11432,10000,8445,6647,4107,1488,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15515,14195,12849,11437,9923,8234,6180,3169,852,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12727,11286,9720,7940,5685,2386,270,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10162,8460,6413,3120,213,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8544,6555,3736,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8459,6466,3650,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10064,8364,6319,3056,208,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12623,11183,9616,7842,5583,2334,262,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15448,14135,12783,11369,9858,8165,6111,3141,850,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18002,16699,15402,14111,12796,11437,10002,8443,6650,4103,1491,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20080,18688,17351,16053,14780,13512,12232,10924,9556,8087,6415,4247,1368,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21678,20106,18659,17300,15997,14731,13483,12241,10990,9699,8348,6873,5151,2547,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22647,20885,19353,17926,16569,15296,14046,12820,11604,10358,9096,7731,6270,4527,1718,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[24181,22111,20357,18840,17424,16079,14820,13587,12383,11193,9973,8750,7423,6001,4311,1559,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22830,21098,19553,18126,16787,15503,14257,13029,11801,10566,9294,7957,6494,4775,2105,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20873,19437,18070,16757,15474,14211,12946,11672,10356,8977,7468,5713,3169,818,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18466,17154,15863,14567,13265,11919,10510,9001,7294,5031,2404,534,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15533,14214,12850,11421,9876,8143,5931,3062,963,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12136,10633,8955,6957,3977,1376,5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8861,6879,3889,726,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6636,3848,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6550,3761,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8758,6779,3809,707,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12041,10531,8858,6860,3909,1344,6,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15472,14152,12780,11358,9812,8081,5871,3044,970,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18463,17157,15860,14574,13262,11918,10516,9000,7295,5037,2414,533,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20865,19427,18060,16746,15466,14202,12941,11665,10348,8968,7463,5709,3162,816,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22832,21097,19555,18132,16791,15501,14249,13025,11802,10564,9292,7959,6501,4774,2102,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24184,22124,20368,18845,17428,16087,14831,13599,12393,11201,9981,8754,7430,6008,4320,1559,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[26519,23585,21531,19799,18287,16891,15567,14327,13116,11934,10770,9586,8391,7103,5734,4105,1426,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24215,22204,20506,18987,17589,16266,14999,13772,12562,11359,10145,8891,7580,6145,4418,1780,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21703,20200,18796,17459,16159,14899,13645,12391,11119,9799,8404,6861,4974,2396,465,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18984,17669,16371,15082,13777,12447,11065,9594,7960,5962,3376,1324,65,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15591,14264,12883,11438,9866,8082,5634,3056,1116,18,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11527,9932,8095,5585,2711,734,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7313,4603,1303,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3950,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3857,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7224,4520,1266,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11422,9830,7997,5493,2665,716,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15531,14205,12827,11376,9809,8019,5600,3038,1112,19,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18984,17668,16371,15084,13779,12450,11064,9595,7967,5962,3378,1326,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21711,20207,18803,17461,16169,14902,13649,12400,11122,9805,8407,6867,4978,2403,469,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24199,22193,20496,18979,17577,16250,14987,13756,12551,11344,10130,8882,7569,6134,4412,1774,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26522,23590,21542,19809,18299,16906,15577,14340,13128,11948,10783,9599,8401,7117,5744,4109,1433,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[30021,25837,22940,20905,19186,17693,16315,15010,13794,12608,11452,10320,9164,8009,6766,5446,3885,1313,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26124,23482,21537,19877,18391,17016,15714,14473,13263,12079,10903,9712,8493,7207,5799,4092,1537,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22627,21041,19584,18208,16895,15620,14371,13126,11882,10615,9297,7885,6299,4174,1874,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19545,18223,16921,15633,14340,13018,11658,10225,8659,6828,4376,2237,648,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15697,14357,12969,11496,9892,8047,5505,3120,1295,133,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10873,9160,7045,4130,1840,331,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5229,1885,116,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[39,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[38,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5134,1844,112,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10768,9051,6940,4061,1801,317,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15651,14305,12918,11450,9850,7989,5477,3121,1297,134,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19553,18220,16935,15633,14334,13020,11656,10221,8658,6825,4374,2238,645,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22618,21031,19571,18199,16890,15616,14370,13126,11883,10613,9295,7885,6299,4173,1871,254,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26126,23483,21538,19879,18392,17009,15710,14465,13257,12071,10893,9701,8480,7197,5786,4088,1529,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30013,25835,22939,20911,19195,17704,16325,15024,13805,12617,11466,10332,9178,8020,6783,5458,3893,1318,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[29422,29294,25117,22259,20248,18551,17076,15717,14440,13246,12087,10959,9855,8738,7619,6428,5162,3670,1217,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28673,25217,22728,20843,19226,17765,16415,15142,13921,12739,11578,10427,9271,8082,6838,5459,3786,1347,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23612,21910,20384,18971,17625,16337,15086,13855,12629,11388,10129,8816,7399,5768,3529,1496,133,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20167,18838,17534,16239,14951,13639,12295,10888,9389,7688,5427,3233,1469,245,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15853,14504,13093,11604,9968,8034,5486,3268,1514,304,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10164,8253,5569,3076,1216,91,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2456,452,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2401,439,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10055,8148,5484,3027,1193,87,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15804,14455,13049,11558,9923,7987,5463,3266,1518,310,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20163,18833,17530,16239,14944,13638,12295,10893,9386,7684,5429,3240,1470,244,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23615,21919,20390,18970,17629,16338,15089,13856,12629,11393,10133,8812,7397,5764,3528,1497,130,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28692,25234,22749,20862,19242,17778,16429,15152,13932,12751,11588,10437,9275,8085,6842,5462,3788,1351,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29395,29266,25098,22241,20234,18538,17065,15709,14430,13239,12081,10958,9855,8738,7617,6429,5162,3671,1222,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[28675,28675,28454,24313,21499,19513,17843,16396,15065,13810,12645,11520,10430,9363,8285,7211,6068,4862,3440,1125,0,0,0,0,0,0,0,0,0,0,0,0],
[29986,27729,24343,21994,20160,18578,17148,15823,14574,13381,12224,11087,9964,8837,7687,6475,5150,3511,1215,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24712,22860,21235,19762,18381,17071,15809,14574,13360,12151,10924,9679,8368,6948,5258,3059,1231,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20851,19510,18190,16901,15616,14307,12985,11610,10157,8552,6542,4308,2428,952,46,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16060,14700,13278,11760,10086,8036,5572,3469,1776,534,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9355,6994,4377,2295,771,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[850,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[820,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9243,6893,4298,2253,752,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16019,14656,13233,11718,10047,7994,5551,3472,1780,536,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20836,19499,18190,16897,15601,14305,12985,11607,10151,8551,6542,4314,2428,948,46,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24727,22870,21248,19770,18394,17080,15814,14580,13362,12149,10928,9678,8368,6949,5261,3063,1240,65,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29987,27725,24342,21985,20158,18581,17150,15823,14576,13381,12221,11087,9968,8844,7688,6476,5150,3515,1217,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28671,28671,28451,24311,21497,19519,17846,16396,15063,13813,12652,11525,10433,9365,8286,7211,6072,4864,3447,1131,0,0,0,0,0,0,0,0,0,0,0,0]],

[[27882,27882,27882,27573,23481,20715,18757,17116,15697,14387,13163,12026,10932,9874,8841,7810,6783,5694,4551,3215,1046,0,0,0,0,0,0,0,0,0,0,0],
[30355,29082,26729,23445,21193,19417,17876,16476,15185,13968,12807,11678,10580,9491,8404,7293,6130,4853,3273,1118,0,0,0,0,0,0,0,0,0,0,0,0],
[26035,23903,22154,20599,19168,17825,16539,15292,14082,12884,11689,10491,9252,7962,6539,4784,2730,1072,34,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21592,20241,18921,17625,16334,15039,13732,12381,10970,9446,7695,5468,3514,1867,619,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16322,14952,13517,11972,10264,8070,5749,3749,2083,817,44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8358,5690,3463,1697,448,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[40,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8244,5604,3402,1662,442,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16284,14897,13460,11930,10206,8037,5727,3739,2092,817,41,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21598,20236,18928,17630,16332,15043,13735,12380,10966,9446,7685,5479,3513,1874,625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26032,23896,22146,20592,19162,17810,16529,15288,14071,12877,11690,10487,9253,7956,6540,4778,2731,1073,34,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30362,29096,26737,23455,21209,19428,17884,16487,15194,13974,12806,11679,10578,9492,8405,7290,6122,4847,3271,1120,0,0,0,0,0,0,0,0,0,0,0,0],
[27875,27875,27875,27565,23478,20711,18762,17117,15695,14387,13167,12034,10937,9879,8849,7815,6787,5702,4558,3223,1056,0,0,0,0,0,0,0,0,0,0,0]],

[[27010,27010,27010,27010,26617,22589,19874,17952,16341,14949,13675,12483,11377,10314,9297,8305,7316,6339,5307,4232,2984,973,0,0,0,0,0,0,0,0,0,0],
[29880,29619,28145,25729,22549,20390,18665,17156,15794,14532,13341,12214,11119,10065,9009,7960,6892,5779,4570,3068,1046,0,0,0,0,0,0,0,0,0,0,0],
[27733,25093,23137,21474,19977,18588,17266,16012,14794,13606,12434,11260,10074,8861,7593,6191,4405,2502,971,26,0,0,0,0,0,0,0,0,0,0,0,0],
[22403,21040,19714,18409,17120,15828,14524,13209,11837,10377,8761,6734,4705,2950,1496,418,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16657,15268,13808,12242,10478,8223,6000,4080,2448,1146,244,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6996,4663,2737,1240,224,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6896,4581,2690,1216,224,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16621,15236,13775,12208,10444,8199,6006,4078,2456,1152,241,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22404,21035,19709,18406,17118,15825,14528,13206,11829,10376,8763,6730,4706,2946,1494,424,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[27740,25091,23135,21478,19975,18585,17266,16006,14787,13600,12422,11256,10069,8855,7586,6189,4402,2501,969,24,0,0,0,0,0,0,0,0,0,0,0,0],
[29875,29614,28145,25734,22553,20392,18665,17161,15797,14532,13346,12209,11118,10054,9004,7955,6885,5773,4565,3063,1043,0,0,0,0,0,0,0,0,0,0,0],
[27006,27006,27006,27006,26615,22592,19875,17958,16349,14955,13680,12492,11389,10327,9308,8318,7330,6349,5320,4242,2990,980,0,0,0,0,0,0,0,0,0,0]],

[[26074,26074,26074,26074,26074,25608,21649,18987,17102,15526,14169,12927,11774,10709,9685,8705,7757,6813,5889,4915,3907,2749,902,0,0,0,0,0,0,0,0,0],
[29272,29272,28746,27142,24697,21630,19552,17875,16410,15082,13848,12695,11590,10533,9510,8500,7496,6480,5419,4277,2885,999,0,0,0,0,0,0,0,0,0,0],
[29179,26648,24234,22405,20807,19352,18001,16714,15487,14294,13135,11983,10840,9686,8490,7253,5894,4144,2367,932,29,0,0,0,0,0,0,0,0,0,0,0],
[23289,21910,20571,19262,17970,16686,15399,14090,12747,11351,9838,8083,6019,4174,2583,1268,320,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17072,15668,14194,12598,10783,8493,6364,4503,2894,1565,554,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5874,3837,2165,884,80,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5779,3774,2125,863,79,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17032,15630,14153,12554,10737,8460,6361,4489,2883,1560,554,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23298,21917,20581,19272,17983,16701,15408,14099,12757,11355,9843,8100,6023,4173,2578,1274,315,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29165,26643,24219,22392,20798,19345,17990,16701,15472,14280,13109,11968,10821,9670,8483,7245,5889,4144,2361,927,28,0,0,0,0,0,0,0,0,0,0,0],
[29282,29282,28761,27153,24703,21639,19560,17885,16421,15090,13859,12702,11596,10546,9515,8502,7499,6476,5419,4273,2881,995,0,0,0,0,0,0,0,0,0,0],
[26068,26068,26068,26068,26068,25604,21647,18986,17110,15536,14175,12936,11778,10715,9692,8713,7767,6824,5896,4929,3922,2756,907,0,0,0,0,0,0,0,0,0]],

[[25072,25072,25072,25072,25072,25072,24542,20664,18057,16215,14677,13354,12152,11034,10005,9024,8087,7186,6292,5419,4509,3575,2510,831,0,0,0,0,0,0,0,0],
[28604,28604,28604,27801,26115,23647,20702,18698,17064,15635,14344,13148,12025,10962,9942,8951,7983,7025,6059,5066,4001,2724,968,0,0,0,0,0,0,0,0,0],
[30364,28139,25563,23400,21675,20135,18726,17403,16153,14952,13779,12647,11530,10409,9292,8157,6955,5648,3978,2293,931,52,0,0,0,0,0,0,0,0,0,0],
[24258,22867,21509,20201,18906,17620,16336,15044,13734,12373,10948,9388,7463,5550,3841,2367,1158,287,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17576,16150,14655,13031,11142,8870,6819,4998,3402,2045,956,197,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4946,3164,1708,613,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4876,3107,1665,592,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17556,16128,14627,13004,11113,8859,6825,5003,3409,2052,960,201,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24244,22852,21505,20186,18900,17612,16330,15040,13722,12373,10943,9378,7464,5543,3841,2367,1156,279,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30366,28150,25573,23409,21684,20145,18733,17414,16158,14949,13789,12646,11528,10416,9299,8157,6960,5647,3982,2296,936,51,0,0,0,0,0,0,0,0,0,0],
[28594,28594,28594,27796,26099,23639,20689,18691,17060,15634,14340,13146,12019,10959,9936,8951,7984,7021,6061,5066,3998,2718,969,0,0,0,0,0,0,0,0,0],
[25070,25070,25070,25070,25070,25070,24542,20663,18058,16223,14685,13361,12157,11039,10010,9028,8094,7193,6301,5426,4520,3584,2513,838,0,0,0,0,0,0,0,0]],

[[24003,24003,24003,24003,24003,24003,24003,23424,19631,17086,15290,13794,12509,11345,10268,9279,8342,7449,6595,5752,4940,4093,3233,2262,767,0,0,0,0,0,0,0],
[27831,27831,27831,27766,26762,25023,22562,19713,17790,16200,14813,13555,12400,11315,10290,9309,8363,7439,6532,5628,4698,3715,2565,952,0,0,0,0,0,0,0,0],
[31363,29396,27114,24575,22605,20959,19473,18102,16804,15589,14414,13281,12171,11089,10009,8929,7826,6684,5456,3899,2294,984,94,0,0,0,0,0,0,0,0,0],
[25283,23875,22517,21188,19895,18603,17331,16050,14764,13444,12079,10629,9018,7066,5272,3675,2284,1149,308,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18199,16763,15241,13585,11593,9392,7397,5608,4020,2636,1478,580,33,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4196,2609,1332,399,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4131,2569,1309,388,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18157,16720,15196,13542,11553,9378,7398,5610,4013,2630,1474,580,33,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[25299,23891,22533,21206,19910,18617,17346,16066,14769,13455,12087,10640,9025,7077,5281,3679,2295,1150,309,0,0,0,0,0,0,0,0,0,0,0,0,0],
[31363,29386,27110,24568,22601,20952,19461,18091,16803,15581,14407,13276,12168,11083,10007,8926,7828,6683,5448,3911,2301,987,98,0,0,0,0,0,0,0,0,0],
[27830,27830,27830,27769,26766,25025,22569,19715,17783,16197,14813,13558,12397,11312,10286,9306,8360,7439,6533,5628,4700,3714,2561,953,0,0,0,0,0,0,0,0],
[23997,23997,23997,23997,23997,23997,23997,23423,19629,17087,15297,13799,12513,11347,10271,9281,8344,7453,6598,5758,4939,4098,3236,2269,771,0,0,0,0,0,0,0]],

[[22876,22876,22876,22876,22876,22876,22876,22876,22260,18555,16074,14327,12875,11632,10511,9473,8529,7635,6790,5988,5201,4441,3665,2881,2011,700,0,0,0,0,0,0],
[26972,26972,26972,26972,26778,25665,23903,21455,18705,16833,15303,13954,12739,11617,10573,9588,8649,7750,6879,6021,5173,4316,3417,2398,949,0,0,0,0,0,0,0],
[32164,30399,28365,26054,23630,21784,20189,18759,17422,16171,14982,13841,12737,11663,10616,9582,8547,7501,6425,5284,3907,2350,1075,167,0,0,0,0,0,0,0,0],
[26409,24982,23611,22279,20970,19686,18410,17142,15867,14582,13263,11894,10437,8767,6895,5187,3670,2340,1243,401,0,0,0,0,0,0,0,0,0,0,0,0],
[18949,17492,15944,14252,12184,10057,8102,6343,4744,3341,2124,1119,369,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3562,2155,1034,239,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3508,2109,1009,232,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18926,17467,15920,14228,12175,10063,8113,6351,4756,3345,2128,1123,372,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26409,24980,23611,22277,20969,19686,18411,17143,15867,14580,13262,11902,10440,8771,6903,5198,3680,2349,1245,410,0,0,0,0,0,0,0,0,0,0,0,0],
[32162,30402,28372,26053,23630,21785,20193,18757,17421,16171,14977,13832,12730,11661,10616,9576,8539,7496,6419,5277,3904,2361,1079,170,0,0,0,0,0,0,0,0],
[26968,26968,26968,26968,26773,25659,23892,21451,18704,16838,15301,13954,12740,11613,10571,9582,8649,7749,6873,6020,5173,4315,3413,2394,953,0,0,0,0,0,0,0],
[22870,22870,22870,22870,22870,22870,22870,22870,22253,18549,16072,14331,12878,11636,10511,9479,8533,7641,6796,5989,5206,4445,3672,2887,2015,708,0,0,0,0,0,0]],

[[21684,21684,21684,21684,21684,21684,21684,21684,21684,21047,17437,15026,13328,11925,10726,9648,8657,7758,6912,6115,5362,4633,3935,3227,2521,1752,635,0,0,0,0,0],
[25999,25999,25999,25999,25999,25666,24491,22712,20295,17648,15828,14341,13042,11862,10783,9784,8841,7948,7096,6279,5481,4696,3909,3105,2204,955,0,0,0,0,0,0],
[32777,31201,29381,27301,24971,22680,20931,19393,18007,16700,15480,14321,13216,12148,11112,10102,9113,8129,7143,6144,5097,3923,2461,1204,279,0,0,0,0,0,0,0],
[27617,26159,24790,23437,22125,20829,19557,18297,17033,15773,14490,13195,11838,10405,8738,6941,5300,3832,2545,1450,581,40,0,0,0,0,0,0,0,0,0,0],
[19863,18385,16816,15079,12977,10897,8992,7237,5645,4203,2931,1835,941,279,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3036,1776,789,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2976,1742,770,122,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19845,18363,16792,15069,12969,10890,8993,7241,5636,4199,2933,1836,940,279,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[27607,26162,24772,23431,22116,20831,19554,18284,17034,15775,14488,13190,11839,10398,8724,6933,5297,3827,2540,1447,580,41,0,0,0,0,0,0,0,0,0,0],
[32783,31213,29380,27317,24977,22693,20935,19401,17997,16703,15486,14326,13217,12153,11119,10108,9115,8134,7155,6152,5099,3931,2467,1207,276,0,0,0,0,0,0,0],
[25992,25992,25992,25992,25992,25663,24481,22713,20297,17635,15833,14340,13043,11866,10785,9783,8843,7946,7089,6276,5476,4693,3912,3102,2208,955,0,0,0,0,0,0],
[21684,21684,21684,21684,21684,21684,21684,21684,21684,21043,17438,15023,13333,11931,10731,9651,8662,7763,6917,6123,5369,4639,3939,3234,2526,1757,643,0,0,0,0,0]],

[[20433,20433,20433,20433,20433,20433,20433,20433,20433,20433,19793,16279,13941,12298,10946,9800,8765,7823,6970,6172,5427,4730,4053,3417,2779,2155,1482,571,0,0,0,0],
[24932,24932,24932,24932,24932,24932,24473,23256,21484,19110,16537,14784,13338,12079,10948,9914,8950,8053,7208,6405,5639,4897,4187,3476,2758,1987,964,1,0,0,0,0],
[33185,31777,30132,28258,26173,23824,21679,19996,18508,17150,15893,14706,13588,12514,11488,10490,9527,8583,7663,6747,5825,4879,3866,2619,1379,417,0,0,0,0,0,0],
[28874,27408,26000,24641,23320,22025,20746,19490,18245,17000,15755,14496,13211,11897,10506,8929,7195,5599,4159,2884,1778,869,213,0,0,0,0,0,0,0,0,0],
[21001,19513,17898,16144,14014,11981,10096,8349,6738,5272,3948,2778,1762,927,301,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2590,1466,598,50,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2544,1433,584,50,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20973,19496,17887,16119,14005,11983,10100,8351,6737,5274,3945,2775,1758,922,295,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28868,27404,25998,24638,23316,22018,20738,19477,18232,16993,15750,14491,13210,11888,10506,8918,7179,5597,4157,2876,1771,868,213,0,0,0,0,0,0,0,0,0],
[33195,31780,30140,28272,26177,23832,21698,20010,18524,17163,15900,14717,13602,12525,11497,10506,9539,8598,7672,6758,5837,4888,3873,2622,1378,419,0,0,0,0,0,0],
[24923,24923,24923,24923,24923,24923,24462,23244,21472,19104,16530,14785,13339,12079,10950,9911,8953,8055,7211,6405,5639,4901,4185,3473,2760,1990,960,0,0,0,0,0],
[20431,20431,20431,20431,20431,20431,20431,20431,20431,20431,19785,16270,13939,12298,10944,9794,8762,7821,6968,6170,5426,4728,4057,3420,2782,2158,1489,574,0,0,0,0]],

[[19130,19130,19130,19130,19130,19130,19130,19130,19130,19130,19130,18489,15076,12815,11231,9929,8831,7847,6953,6150,5403,4710,4069,3454,2881,2315,1772,1208,479,0,0,0],
[23730,23730,23730,23730,23730,23730,23730,23170,21942,20189,17863,15369,13677,12285,11069,9978,8990,8073,7224,6428,5679,4967,4290,3640,3011,2384,1738,958,21,0,0,0],
[33385,32107,30628,28936,27040,24930,22615,20594,18965,17522,16205,14980,13835,12745,11715,10724,9776,8856,7976,7109,6264,5425,4587,3721,2768,1587,602,7,0,0,0,0],
[30175,28693,27267,25898,24559,23254,21978,20721,19479,18246,17020,15802,14573,13333,12068,10747,9329,7670,6101,4678,3385,2260,1299,538,49,0,0,0,0,0,0,0],
[22431,20906,19276,17453,15337,13337,11472,9730,8107,6609,5230,3986,2875,1899,1089,447,42,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2205,1206,444,5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2160,1178,431,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22405,20885,19261,17445,15335,13337,11470,9723,8105,6606,5227,3982,2872,1897,1087,447,43,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30183,28702,27272,25895,24561,23254,21976,20715,19478,18250,17030,15809,14575,13342,12071,10750,9328,7672,6101,4671,3386,2257,1297,539,51,0,0,0,0,0,0,0],
[33371,32093,30611,28922,27032,24933,22605,20591,18963,17513,16201,14972,13827,12743,11711,10719,9772,8858,7970,7108,6264,5425,4580,3717,2758,1580,602,7,0,0,0,0],
[23750,23750,23750,23750,23750,23750,23750,23188,21948,20201,17877,15384,13689,12295,11077,9987,8996,8082,7232,6434,5681,4974,4297,3648,3015,2387,1740,962,21,0,0,0],
[19113,19113,19113,19113,19113,19113,19113,19113,19113,19113,19113,18472,15068,12809,11231,9929,8827,7845,6950,6150,5402,4712,4068,3455,2882,2320,1775,1213,486,0,0,0]],

[[17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17136,13834,11652,10130,8885,7839,6907,6068,5316,4622,3984,3398,2845,2337,1843,1385,927,372,0,0],
[22439,22439,22439,22439,22439,22439,22439,22439,21799,20579,18851,16593,14166,12522,11181,10007,8968,8020,7154,6352,5605,4908,4256,3640,3056,2498,1967,1431,853,51,0,0],
[33314,32165,30819,29283,27575,25676,23568,21257,19361,17783,16387,15103,13914,12806,11757,10763,9815,8912,8050,7217,6419,5643,4891,4147,3416,2665,1813,821,114,0,0,0],
[31433,29925,28483,27091,25741,24427,23141,21882,20651,19428,18220,17027,15842,14660,13471,12268,11044,9769,8380,6824,5391,4088,2925,1903,1048,385,5,0,0,0,0,0],
[24258,22720,21068,19175,17086,15119,13265,11511,9873,8336,6914,5597,4397,3305,2344,1510,819,295,2,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1884,989,319,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1848,964,312,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24239,22690,21042,19168,17082,15111,13258,11509,9870,8332,6913,5595,4396,3308,2343,1510,819,293,2,0,0,0,0,0,0,0,0,0,0,0,0,0],
[31440,29932,28490,27097,25740,24430,23143,21884,20646,19430,18228,17032,15841,14658,13473,12268,11046,9767,8380,6824,5393,4093,2924,1906,1047,385,6,0,0,0,0,0],
[33321,32167,30823,29291,27579,25674,23579,21272,19362,17791,16385,15102,13915,12801,11754,10760,9824,8917,8050,7224,6417,5639,4886,4147,3409,2659,1812,817,114,0,0,0],
[22432,22432,22432,22432,22432,22432,22432,22432,21800,20564,18854,16582,14167,12524,11173,10005,8961,8016,7150,6346,5600,4908,4258,3642,3062,2503,1967,1431,854,51,0,0],
[17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17755,17128,13836,11654,10138,8894,7843,6912,6070,5320,4628,3991,3403,2850,2340,1852,1393,931,377,0,0]],

[[16346,16346,16346,16346,16346,16346,16346,16346,16346,16346,16346,16346,16346,15748,12567,10466,9009,7823,6825,5953,5165,4465,3825,3243,2717,2224,1785,1366,991,636,249,0],
[20989,20989,20989,20989,20989,20989,20989,20989,20980,20319,19107,17431,15236,12889,11300,9999,8875,7879,6984,6162,5411,4721,4080,3487,2938,2425,1945,1494,1071,653,87,0],
[32956,31920,30698,29307,27749,26026,24122,22043,19779,17940,16411,15045,13798,12651,11577,10566,9617,8714,7865,7055,6286,5558,4854,4182,3534,2917,2313,1712,1056,290,0,0],
[32517,30983,29516,28111,26739,25414,24121,22855,21614,20399,19211,18039,16878,15730,14601,13474,12357,11233,10094,8943,7731,6356,5020,3802,2717,1774,978,359,5,0,0,0],
[26699,25134,23454,21517,19461,17503,15658,13901,12239,10669,9196,7815,6532,5341,4248,3267,2383,1608,954,432,79,0,0,0,0,0,0,0,0,0,0,0],
[1618,814,228,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1579,789,217,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26683,25112,23432,21502,19460,17506,15647,13893,12232,10664,9185,7810,6526,5337,4248,3265,2382,1605,953,432,79,0,0,0,0,0,0,0,0,0,0,0],
[32511,30979,29515,28106,26738,25410,24114,22854,21617,20400,19210,18030,16873,15731,14594,13463,12345,11223,10090,8935,7725,6352,5019,3806,2716,1772,977,358,4,0,0,0],
[32958,31919,30702,29310,27750,26027,24130,22047,19772,17946,16412,15041,13802,12654,11581,10572,9620,8718,7863,7053,6290,5555,4852,4179,3539,2916,2314,1712,1053,291,0,0],
[20992,20992,20992,20992,20992,20992,20992,20992,20981,20321,19111,17437,15234,12892,11299,10000,8878,7882,6982,6166,5412,4720,4075,3484,2931,2419,1941,1495,1073,651,87,0],
[16349,16349,16349,16349,16349,16349,16349,16349,16349,16349,16349,16349,16349,15748,12566,10473,9020,7831,6838,5961,5170,4475,3837,3255,2725,2236,1792,1377,1000,642,255,0]],

[[14883,14883,14883,14883,14883,14883,14883,14883,14883,14883,14883,14883,14883,14883,14326,11264,9250,7859,6738,5800,4979,4247,3603,3020,2494,2027,1598,1224,880,590,336,112],
[19442,19442,19442,19442,19442,19442,19442,19442,19442,19426,18773,17592,15966,13842,11560,10024,8769,7700,6746,5901,5132,4435,3795,3209,2677,2187,1746,1346,984,664,381,105],
[32232,31293,30189,28932,27512,25946,24214,22331,20279,18057,16278,14786,13454,12246,11128,10095,9129,8221,7374,6567,5818,5106,4438,3813,3217,2669,2154,1673,1230,820,431,9],
[32975,31425,29926,28495,27110,25768,24458,23186,21950,20735,19553,18387,17253,16137,15042,13970,12917,11875,10857,9849,8852,7876,6895,5915,4901,3796,2758,1849,1078,469,62,0],
[30239,28651,26942,24966,22942,20995,19144,17372,15682,14073,12548,11108,9738,8462,7258,6141,5103,4151,3274,2493,1796,1195,691,298,42,0,0,0,0,0,0,0],
[1381,663,155,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1351,647,152,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30226,28641,26934,24970,22942,21002,19149,17377,15688,14076,12558,11113,9745,8467,7261,6147,5105,4155,3285,2494,1798,1193,693,298,42,0,0,0,0,0,0,0],
[32974,31413,29921,28500,27111,25758,24460,23177,21945,20729,19542,18383,17244,16131,15031,13953,12902,11867,10843,9845,8849,7866,6889,5908,4901,3788,2755,1846,1077,469,62,0],
[32247,31314,30200,28936,27525,25955,24227,22351,20293,18071,16291,14803,13466,12254,11145,10111,9144,8235,7387,6581,5827,5117,4451,3819,3229,2679,2161,1678,1237,825,437,9],
[19418,19418,19418,19418,19418,19418,19418,19418,19418,19402,18748,17569,15952,13823,11543,10004,8751,7680,6732,5889,5120,4421,3779,3196,2666,2176,1737,1335,973,652,370,108],
[14893,14893,14893,14893,14893,14893,14893,14893,14893,14893,14893,14893,14893,14893,14333,11271,9258,7876,6751,5812,4987,4258,3615,3031,2508,2037,1609,1233,893,599,346,117]],

[[13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,13417,12912,9970,8048,6729,5669,4791,4027,3353,2764,2240,1772,1365,1003,695,426,218,68],
[17781,17781,17781,17781,17781,17781,17781,17781,17781,17781,17772,17155,16025,14465,12422,10200,8715,7515,6491,5596,4796,4081,3434,2853,2326,1854,1430,1056,731,455,234,74],
[31225,30379,29377,28231,26943,25525,23952,22244,20392,18376,16201,14448,12991,11685,10518,9440,8445,7516,6658,5852,5109,4407,3762,3156,2603,2085,1622,1204,835,517,260,76],
[30693,29118,27588,26132,24740,23374,22049,20760,19518,18303,17119,15961,14838,13737,12669,11631,10618,9638,8687,7763,6865,6006,5186,4403,3654,2950,2297,1699,1167,709,346,95],
[36672,35052,33321,31327,29315,27375,25512,23723,22005,20355,18779,17272,15828,14461,13153,11918,10745,9645,8598,7619,6696,5842,5045,4312,3625,3003,2424,1904,1417,968,558,160],
[1316,672,225,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123,123],
[51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51,51],
[33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33],
[39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39],
[46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46,46],
[1293,663,228,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124],
[36665,35051,33313,31332,29320,27386,25519,23732,22009,20363,18784,17279,15837,14463,13162,11925,10745,9640,8601,7619,6698,5849,5046,4313,3629,3005,2429,1900,1422,974,562,164],
[30695,29108,27597,26140,24735,23362,22056,20776,19515,18303,17121,15959,14839,13737,12673,11631,10625,9637,8687,7768,6875,6012,5185,4400,3654,2955,2295,1698,1167,706,345,95],
[31220,30377,29372,28227,26937,25513,23946,22241,20387,18377,16197,14444,12983,11695,10516,9440,8449,7523,6659,5859,5115,4414,3764,3159,2601,2087,1620,1204,835,520,263,81],
[17778,17778,17778,17778,17778,17778,17778,17778,17778,17778,17768,17148,16019,14462,12420,10194,8707,7509,6481,5585,4791,4075,3429,2848,2323,1852,1429,1056,726,453,229,68],
[13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,13411,12904,9966,8045,6732,5672,4790,4025,3353,2766,2241,1774,1366,1003,695,429,220,71]],

[[12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,12118,11665,8856,7026,5782,4785,3973,3268,2650,2121,1652,1246,901,601,361,169,47],
[16292,16292,16292,16292,16292,16292,16292,16292,16292,16292,16292,16292,15739,14678,13191,11243,9085,7648,6497,5527,4683,3937,3274,2684,2162,1697,1286,931,627,375,181,51],
[30537,29774,28868,27834,26659,25363,23942,22396,20716,18902,16939,14819,13067,11626,10359,9223,8182,7225,6345,5532,4782,4084,3446,2856,2315,1827,1393,1004,676,403,188,52],
[29416,27803,26263,24777,23351,21970,20629,19327,18069,16847,15656,14502,13381,12301,11247,10227,9247,8303,7388,6523,5686,4901,4155,3457,2808,2215,1673,1195,789,459,208,57],
[40063,38427,36669,34664,32668,30724,28850,27031,25284,23597,21963,20389,18875,17414,16011,14666,13369,12123,10931,9777,8684,7624,6618,5642,4712,3806,2933,2089,1259,626,262,59],
[2102,1525,1131,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,1057,869,384,67],
[409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,409,80],
[275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,275,79],
[283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,283,82],
[406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,406,75],
[2075,1510,1131,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,1056,869,384,66],
[40063,38423,36666,34665,32667,30731,28851,27043,25285,23600,21971,20397,18882,17414,16005,14663,13364,12117,10926,9775,8679,7621,6616,5641,4712,3804,2935,2088,1263,622,265,63],
[29420,27809,26269,24788,23358,21977,20637,19340,18077,16861,15663,14512,13389,12296,11254,10239,9252,8309,7394,6528,5693,4905,4158,3457,2808,2215,1675,1200,792,462,208,54],
[30523,29763,28851,27809,26639,25345,23926,22375,20694,18882,16917,14797,13051,11622,10359,9216,8178,7226,6343,5528,4778,4078,3445,2855,2317,1828,1391,1007,676,401,192,52],
[16299,16299,16299,16299,16299,16299,16299,16299,16299,16299,16299,16299,15750,14678,13195,11239,9079,7640,6488,5520,4677,3935,3274,2682,2158,1695,1287,927,623,371,177,48],
[12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,12112,11664,8854,7026,5788,4787,3974,3266,2650,2122,1653,1249,900,603,362,174,49]],

[[10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10904,10517,7844,6112,4941,4017,3265,2620,2064,1595,1187,843,558,324,156,37],
[14904,14904,14904,14904,14904,14904,14904,14904,14904,14904,14904,14904,14904,14460,13469,12058,10199,8100,6713,5615,4698,3907,3217,2612,2079,1614,1206,857,566,330,152,42],
[29894,29205,28384,27446,26387,25209,23922,22531,21005,19364,17593,15694,13645,11874,10461,9232,8124,7127,6217,5380,4612,3914,3271,2686,2163,1690,1274,907,595,350,161,42],
[28927,27282,25716,24204,22748,21342,19981,18662,17384,16147,14949,13783,12668,11594,10547,9545,8577,7655,6777,5942,5145,4405,3701,3056,2459,1920,1439,1020,671,385,180,54],
[41436,39773,37994,35990,33994,32052,30153,28311,26524,24782,23097,21455,19867,18311,16817,15352,13944,12558,11232,9930,8672,7435,6232,5062,3913,2787,1938,1293,797,433,179,36],
[3584,3063,2719,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2660,2654,2388,1918,1280,561,216,48],
[953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,953,838,256,54],
[646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,646,278,53],
[656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,656,278,53],
[944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,944,829,247,48],
[3570,3062,2728,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2665,2659,2391,1925,1284,566,221,52],
[41421,39757,37981,35988,33983,32044,30153,28308,26521,24780,23090,21455,19859,18311,16815,15358,13943,12565,11231,9926,8667,7438,6237,5063,3911,2789,1940,1292,793,438,181,37],
[28921,27277,25714,24204,22746,21337,19975,18655,17376,16143,14945,13785,12661,11586,10533,9535,8574,7644,6771,5935,5135,4395,3692,3046,2452,1912,1431,1016,668,381,179,52],
[29906,29215,28395,27456,26400,25224,23938,22535,21019,19372,17606,15702,13654,11884,10477,9239,8131,7133,6226,5388,4622,3921,3279,2697,2168,1692,1272,911,603,351,166,46],
[14899,14899,14899,14899,14899,14899,14899,14899,14899,14899,14899,14899,14899,14450,13461,12051,10207,8099,6706,5615,4691,3902,3215,2608,2082,1614,1207,863,570,334,156,43],
[10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10897,10511,7842,6109,4948,4020,3263,2619,2060,1593,1189,841,558,324,152,38]],

[[9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9757,9443,6909,5276,4186,3331,2646,2064,1569,1159,812,529,308,141,39],
[13530,13530,13530,13530,13530,13530,13530,13530,13530,13530,13530,13530,13530,13530,13209,12313,10993,9249,7199,5857,4820,3960,3223,2592,2048,1573,1165,822,536,309,141,37],
[29246,28620,27883,27039,26088,25027,23868,22601,21230,19749,18160,16457,14620,12650,10848,9452,8246,7171,6211,5334,4547,3834,3184,2597,2074,1604,1196,847,554,317,145,37],
[28764,27103,25501,23959,22470,21034,19636,18299,17000,15742,14529,13359,12234,11153,10104,9108,8157,7250,6384,5563,4797,4081,3415,2796,2234,1728,1285,899,576,318,136,31],
[41450,39770,37985,36005,33996,32040,30126,28252,26424,24640,22880,21178,19502,17872,16271,14704,13163,11674,10197,8742,7323,5926,4711,3722,2886,2172,1569,1077,685,393,179,49],
[5757,5286,4968,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4899,4730,4357,3844,3205,2436,1527,816,414,171,41],
[1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1645,1234,493,186,41],
[1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,1114,560,193,43],
[1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,1122,562,194,44],
[1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1640,1233,490,185,42],
[5736,5277,4971,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4903,4734,4363,3848,3206,2443,1531,823,419,177,42],
[41447,39762,37977,35997,33990,32036,30121,28244,26421,24630,22877,21172,19491,17863,16260,14702,13160,11665,10188,8743,7316,5923,4709,3720,2881,2162,1566,1074,682,390,176,47],
[28771,27102,25505,23965,22477,21042,19650,18304,17006,15749,14536,13362,12239,11152,10112,9111,8161,7242,6382,5566,4795,4075,3415,2794,2234,1726,1285,897,578,320,138,33],
[29251,28628,27897,27045,26091,25033,23872,22603,21237,19762,18164,16462,14634,12660,10860,9461,8250,7177,6216,5348,4553,3839,3188,2600,2077,1609,1199,848,555,319,144,37],
[13523,13523,13523,13523,13523,13523,13523,13523,13523,13523,13523,13523,13523,13523,13203,12302,10987,9236,7200,5853,4821,3956,3226,2594,2045,1572,1166,821,533,310,144,39],
[9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9752,9426,6903,5269,4187,3331,2643,2063,1566,1156,810,528,305,139,36]],

[[8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8664,8415,6036,4508,3501,2716,2094,1579,1146,799,512,293,133,32],
[12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,12176,11999,11201,9992,8364,6380,5091,4102,3297,2624,2054,1562,1149,802,517,296,136,36],
[28536,27974,27314,26546,25693,24743,23702,22566,21335,20010,18583,17050,15414,13664,11786,9941,8547,7362,6322,5393,4569,3826,3156,2560,2030,1564,1157,810,526,299,134,34],
[28893,27187,25564,23985,22463,20999,19572,18205,16874,15596,14360,13176,12035,10939,9889,8888,7934,7030,6172,5364,4608,3908,3256,2660,2114,1638,1207,849,547,304,132,30],
[40291,38595,36801,34861,32848,30870,28927,27017,25139,23296,21490,19706,17949,16218,14526,12847,11195,9586,8180,6954,5840,4849,3958,3179,2481,1883,1370,945,597,342,154,42],
[8617,8186,7885,7784,7784,7784,7784,7784,7784,7784,7784,7784,7784,7784,7784,7784,7784,7770,7553,7195,6723,6147,5473,4697,3822,2823,1769,1119,671,361,150,35],
[2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2459,2367,1642,790,392,162,40],
[1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,1664,913,412,164,37],
[1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,1672,916,412,165,33],
[2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2451,2364,1637,791,391,160,39],
[8610,8192,7894,7789,7789,7789,7789,7789,7789,7789,7789,7789,7789,7789,7789,7789,7789,7776,7561,7202,6728,6156,5479,4702,3825,2824,1770,1118,670,352,152,37],
[40289,38579,36802,34847,32851,30865,28926,27018,25144,23296,21484,19694,17951,16214,14524,12840,11200,9574,8182,6949,5845,4846,3968,3173,2485,1882,1369,944,595,340,155,42],
[28895,27189,25563,23990,22457,21002,19575,18207,16876,15601,14368,13176,12035,10942,9894,8892,7939,7032,6175,5367,4612,3906,3261,2662,2119,1638,1212,848,548,302,132,29],
[28528,27964,27306,26540,25688,24739,23700,22562,21330,20000,18576,17050,15411,13658,11781,9940,8548,7360,6313,5390,4569,3822,3154,2562,2028,1561,1158,809,525,300,136,34],
[12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,12175,11992,11204,9992,8349,6382,5091,4105,3303,2625,2051,1564,1152,803,519,297,136,37],
[8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8660,8404,6027,4504,3501,2717,2091,1575,1144,799,516,295,134,34]],

[[7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7622,7435,5215,3797,2875,2164,1616,1164,796,509,289,128,34],
[10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10780,10128,9035,7521,5621,4380,3452,2704,2091,1573,1149,795,511,286,130,33],
[27678,27170,26579,25894,25139,24292,23368,22359,21265,20085,18817,17454,15998,14447,12785,11003,9146,7730,6558,5552,4659,3878,3186,2568,2024,1549,1139,797,512,290,130,33],
[29226,27505,25833,24229,22663,21158,19699,18294,16927,15621,14354,13147,11975,10875,9796,8791,7828,6916,6063,5248,4501,3796,3159,2575,2042,1571,1159,808,513,291,134,33],
[38016,36298,34524,32666,30686,28693,26715,24765,22832,20940,19052,17195,15384,13752,12246,10844,9541,8340,7210,6185,5235,4372,3595,2890,2276,1733,1266,880,566,317,136,33],
[12277,11872,11575,11371,11329,11329,11329,11329,11329,11329,11329,11329,11298,11102,10810,10424,9951,9399,8766,8050,7254,6372,5395,4319,3124,2124,1458,964,593,324,140,36],
[3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3377,3018,2068,1138,654,343,145,35],
[2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,2287,1336,701,356,146,34],
[2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,2294,1334,702,356,143,35],
[3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3371,3015,2062,1139,652,343,146,34],
[12263,11870,11574,11371,11332,11332,11332,11332,11332,11332,11332,11332,11303,11106,10812,10429,9957,9404,8771,8053,7261,6376,5399,4326,3129,2127,1463,965,593,324,138,36],
[38024,36308,34530,32676,30688,28688,26710,24768,22838,20934,19052,17199,15383,13755,12249,10848,9539,8331,7210,6179,5234,4371,3592,2890,2272,1733,1265,878,561,315,136,34],
[29217,27509,25842,24228,22672,21164,19699,18299,16935,15616,14365,13147,11984,10867,9803,8795,7824,6920,6054,5255,4496,3806,3158,2574,2039,1573,1159,810,515,294,132,33],
[27677,27170,26576,25893,25135,24290,23368,22357,21259,20083,18816,17453,16004,14446,12784,10999,9147,7728,6562,5554,4665,3879,3183,2565,2020,1549,1140,792,511,289,132,34],
[10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10839,10779,10128,9031,7526,5619,4376,3451,2704,2089,1577,1146,791,510,289,132,32],
[7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7613,7428,5211,3794,2877,2166,1615,1164,795,508,286,127,34]],

[[6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6630,6506,4450,3143,2306,1676,1198,814,510,288,128,33],
[9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9523,9070,8107,6735,4913,3724,2860,2175,1618,1167,798,509,281,126,31],
[26648,26182,25653,25051,24386,23647,22837,21953,20994,19957,18838,17644,16361,14989,13530,11960,10286,8487,6997,5841,4860,4003,3262,2610,2044,1557,1141,791,502,285,127,32],
[29775,28039,26333,24672,23083,21522,20013,18569,17152,15808,14510,13264,12069,10928,9842,8811,7825,6910,6033,5226,4466,3770,3126,2543,2014,1552,1148,800,511,283,126,31],
[34228,32515,30756,28968,27149,25346,23537,21768,20054,18392,16794,15273,13818,12446,11139,9901,8751,7662,6657,5716,4854,4064,3344,2699,2129,1618,1181,818,525,296,133,33],
[17182,16802,16480,16211,15972,15757,15552,15319,15060,14748,14396,13972,13490,12950,12338,11661,10917,10103,9205,8236,7176,6036,4774,3464,2546,1860,1320,893,559,305,134,35],
[4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4372,4303,3656,2509,1528,971,583,316,135,32],
[2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,2980,1817,1054,614,322,136,34],
[2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,2984,1816,1053,613,319,138,31],
[4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4377,4305,3658,2509,1529,970,583,312,135,33],
[17182,16814,16491,16215,15982,15758,15541,15307,15051,14742,14383,13963,13485,12939,12336,11660,10914,10097,9205,8233,7174,6030,4774,3461,2550,1865,1323,894,560,312,135,34],
[34226,32497,30733,28956,27143,25335,23544,21779,20054,18394,16804,15278,13824,12445,11131,9902,8744,7666,6652,5718,4853,4064,3344,2699,2125,1620,1181,818,523,294,132,33],
[29773,28032,26324,24671,23066,21516,20011,18558,17156,15800,14497,13259,12062,10920,9838,8797,7827,6899,6034,5221,4459,3765,3117,2542,2011,1549,1137,798,509,284,123,30],
[26657,26191,25663,25063,24397,23665,22846,21966,21005,19967,18850,17653,16369,15002,13538,11972,10297,8503,7004,5847,4854,4005,3266,2610,2053,1559,1139,790,507,285,127,32],
[9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9525,9072,8112,6733,4922,3729,2863,2176,1624,1173,806,514,288,130,32],
[6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6616,6493,4438,3135,2307,1675,1193,806,502,283,126,33]],

[[5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5682,5617,3742,2550,1804,1251,843,526,289,127,31],
[8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8232,8025,7215,5980,4268,3121,2318,1700,1211,820,520,287,129,35],
[25343,24913,24440,23919,23347,22711,22017,21264,20439,19551,18587,17554,16444,15256,13977,12619,11166,9608,7938,6347,5180,4221,3404,2703,2104,1587,1151,798,509,285,126,31],
[30591,28824,27098,25397,23745,22142,20575,19065,17617,16198,14856,13563,12325,11147,10020,8963,7944,6999,6106,5285,4507,3794,3135,2552,2027,1559,1155,807,520,292,134,33],
[32119,30391,28662,26928,25217,23510,21858,20239,18668,17159,15711,14312,12975,11712,10499,9358,8282,7276,6322,5446,4628,3892,3209,2592,2032,1555,1131,779,496,278,119,29],
[20219,19849,19489,19152,18795,18434,18035,17608,17128,16600,16027,15390,14690,13934,13110,12210,11248,10205,9085,7876,6576,5161,3907,3015,2296,1711,1227,837,527,293,130,32],
[5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5455,5159,4312,2981,1975,1341,888,546,301,129,32],
[3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,3729,2351,1465,931,558,300,130,32],
[3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,3733,2348,1463,932,555,303,129,31],
[5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5465,5162,4322,2984,1981,1347,890,552,301,133,34],
[20200,19831,19479,19141,18791,18429,18027,17603,17122,16598,16017,15382,14685,13925,13100,12213,11244,10199,9082,7871,6569,5155,3908,3006,2295,1706,1229,832,522,293,126,32],
[32129,30402,28671,26936,25214,23525,21860,20233,18680,17158,15705,14314,12980,11701,10504,9360,8287,7274,6329,5455,4637,3889,3210,2595,2044,1565,1145,789,498,278,122,27],
[30591,28825,27082,25393,23746,22128,20567,19063,17606,16205,14852,13559,12324,11143,10018,8951,7941,6996,6104,5271,4502,3788,3141,2548,2020,1548,1143,799,520,294,132,36],
[25343,24911,24439,23918,23346,22714,22016,21263,20439,19546,18586,17554,16441,15249,13978,12620,11167,9612,7939,6350,5181,4223,3409,2710,2104,1593,1155,801,511,284,126,30],
[8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8222,8010,7211,5967,4257,3113,2312,1695,1202,810,509,278,123,29],
[5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5683,5618,3736,2552,1812,1257,845,529,294,132,33]],

[[4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4792,4771,3075,2008,1351,881,544,297,125,30],
[6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6978,6944,6330,5250,3669,2566,1835,1281,859,538,297,133,35],
[23672,23256,22837,22389,21914,21392,20824,20206,19530,18790,17990,17121,16191,15182,14096,12939,11701,10362,8942,7410,5783,4585,3640,2862,2203,1654,1188,814,515,289,127,30],
[31820,30037,28265,26515,24796,23110,21474,19895,18358,16875,15459,14095,12786,11550,10367,9252,8196,7200,6274,5403,4607,3865,3189,2579,2034,1548,1131,772,490,270,118,26],
[30633,28923,27224,25548,23900,22292,20718,19203,17733,16322,14952,13642,12389,11197,10056,8982,7967,7008,6111,5277,4501,3785,3135,2544,2019,1555,1151,810,527,298,137,39],
[22354,21970,21582,21164,20723,20235,19703,19127,18496,17802,17054,16242,15363,14420,13405,12315,11146,9895,8548,7106,5572,4439,3532,2785,2155,1616,1172,810,513,287,128,29],
[6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6613,6580,6007,4998,3509,2473,1770,1244,836,523,288,125,33],
[4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4536,4519,2938,1932,1313,861,531,291,129,33],
[4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4538,4520,2934,1931,1311,855,530,289,126,30],
[6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6607,6576,6005,4996,3512,2472,1776,1249,838,526,291,129,33],
[22362,21974,21588,21168,20723,20238,19709,19128,18499,17806,17058,16248,15370,14426,13415,12321,11151,9903,8557,7113,5581,4441,3541,2786,2154,1619,1174,804,511,287,125,31],
[30939,29222,27516,25825,24169,22555,20971,19438,17953,16528,15150,13829,12565,11356,10211,9121,8094,7120,6218,5369,4580,3857,3193,2596,2052,1580,1167,812,527,299,138,36],
[31527,29742,27979,26234,24523,22857,21236,19662,18138,16668,15263,13911,12616,11384,10222,9111,8064,7093,6171,5309,4526,3795,3132,2533,1991,1525,1115,770,488,271,117,27],
[23665,23263,22839,22393,21916,21394,20828,20209,19536,18789,17991,17124,16189,15179,14100,12940,11698,10367,8940,7406,5784,4581,3638,2861,2199,1650,1193,815,518,289,129,31],
[6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6979,6942,6334,5255,3668,2562,1830,1278,854,531,291,127,33],
[4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4785,4764,3071,2008,1359,887,547,300,129,33]],

[[3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,3947,2464,1525,962,573,311,135,30],
[5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5781,5459,4557,3131,2060,1395,917,565,308,134,35],
[21443,21044,20668,20301,19930,19533,19109,18644,18133,17570,16951,16274,15524,14716,13832,12879,11852,10747,9548,8268,6881,5378,4051,3117,2370,1756,1260,851,537,299,130,33],
[33202,31408,29609,27800,26012,24261,22534,20856,19231,17661,16155,14712,13322,12015,10775,9592,8483,7440,6473,5572,4730,3970,3268,2645,2084,1589,1159,802,513,286,127,32],
[29856,28157,26481,24840,23246,21683,20181,18713,17296,15934,14627,13359,12157,11000,9900,8853,7866,6934,6058,5235,4479,3772,3130,2541,2012,1545,1142,798,513,286,125,31],
[23991,23587,23151,22660,22127,21540,20890,20186,19414,18576,17680,16709,15664,14543,13345,12065,10698,9230,7647,6146,5035,4117,3331,2659,2072,1574,1150,793,508,286,128,33],
[7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7815,7614,6859,5705,4097,3008,2250,1659,1185,805,506,283,125,31],
[5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5393,5333,3585,2464,1754,1225,825,519,290,133,32],
[5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5387,5328,3572,2458,1747,1216,820,512,284,124,29],
[7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7816,7621,6870,5713,4108,3016,2253,1663,1189,810,515,289,129,34],
[23997,23593,23155,22672,22132,21546,20901,20191,19422,18584,17682,16716,15669,14547,13349,12072,10705,9238,7660,6154,5046,4120,3336,2654,2072,1570,1148,792,506,288,129,33],
[29862,28159,26493,24855,23256,21704,20182,18725,17311,15941,14624,13374,12159,11004,9907,8868,7875,6932,6068,5246,4477,3775,3128,2550,2021,1557,1150,799,513,287,128,30],
[33180,31399,29589,27800,26015,24247,22526,20854,19224,17662,16151,14700,13328,12018,10767,9586,8483,7444,6464,5558,4735,3970,3270,2639,2074,1582,1156,800,515,286,128,31],
[21462,21063,20683,20308,19927,19534,19110,18646,18136,17572,16954,16274,15523,14713,13832,12882,11852,10741,9549,8267,6881,5382,4055,3119,2368,1757,1261,854,535,298,132,34],
[5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5784,5464,4558,3131,2061,1397,918,561,307,134,34],
[3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,3935,2457,1517,964,575,309,132,30]],

[[3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,3160,1910,1097,634,329,142,34],
[4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4643,4567,3872,2641,1601,1012,608,325,139,34],
[18238,17834,17489,17195,16939,16713,16491,16245,15960,15629,15239,14794,14275,13697,13053,12321,11531,10653,9701,8664,7546,6318,4983,3595,2641,1921,1358,915,570,315,138,36],
[35583,33794,31965,30100,28220,26323,24435,22603,20813,19083,17423,15827,14324,12892,11533,10252,9055,7930,6880,5918,5022,4204,3472,2801,2207,1688,1238,863,555,319,142,40],
[29198,27514,25869,24276,22710,21213,19749,18341,16966,15656,14378,13164,11985,10870,9796,8778,7808,6898,6025,5227,4472,3777,3131,2548,2019,1555,1144,792,501,277,120,24],
[25277,24851,24354,23795,23175,22484,21726,20896,20002,19022,17979,16849,15643,14354,12971,11501,9915,8218,6807,5700,4750,3927,3200,2578,2021,1542,1132,788,505,284,126,32],
[9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,9064,8642,7743,6455,4752,3624,2795,2143,1605,1167,805,518,296,135,37],
[6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6292,6179,4261,3036,2239,1633,1164,793,500,280,121,30],
[6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6304,6190,4271,3042,2247,1639,1173,798,505,285,124,30],
[9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,9061,8636,7745,6446,4744,3613,2790,2134,1596,1156,794,511,285,130,34],
[25275,24841,24349,23791,23167,22477,21720,20890,19993,19020,17974,16843,15640,14353,12969,11496,9913,8219,6808,5691,4747,3931,3208,2574,2022,1545,1133,787,503,286,129,32],
[29201,27525,25880,24280,22726,21222,19748,18348,16971,15659,14388,13161,11993,10867,9798,8779,7811,6901,6034,5232,4472,3775,3136,2546,2017,1559,1145,791,502,279,119,27],
[35579,33787,31958,30099,28221,26327,24455,22606,20811,19081,17422,15838,14330,12892,11538,10256,9058,7933,6888,5919,5031,4214,3463,2808,2210,1682,1238,867,563,324,147,42],
[18242,17837,17494,17197,16945,16712,16480,16233,15950,15620,15233,14785,14271,13688,13042,12314,11520,10648,9694,8660,7533,6308,4980,3590,2631,1914,1351,907,559,306,133,32],
[4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4645,4569,3878,2646,1605,1015,611,329,140,37],
[3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,3151,1901,1093,637,329,140,33]],

[[2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,2434,1410,732,368,149,37],
[3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3587,3200,2185,1188,681,355,149,35],
[13043,12626,12302,12087,12044,12044,12044,12044,12044,12044,12044,12044,12010,11806,11487,11072,10568,9974,9300,8535,7683,6740,5703,4555,3277,2223,1525,1006,620,336,145,36],
[39690,37899,36045,34111,32042,29949,27881,25832,23815,21813,19840,17894,15995,14286,12708,11240,9889,8626,7451,6378,5398,4501,3687,2970,2326,1762,1289,882,553,302,132,31],
[28798,27140,25525,23967,22452,20991,19569,18194,16861,15578,14338,13148,12001,10902,9850,8846,7881,6983,6121,5320,4562,3867,3220,2630,2091,1617,1199,839,552,319,144,39],
[26353,25880,25325,24693,23982,23195,22322,21377,20347,19244,18053,16765,15393,13927,12357,10672,8916,7575,6446,5465,4611,3843,3166,2557,2027,1557,1151,802,518,295,134,33],
[10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10325,10271,9663,8638,7223,5439,4264,3381,2659,2067,1564,1144,794,513,292,132,34],
[7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7263,7088,5020,3685,2803,2126,1588,1151,796,511,287,130,34],
[7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7268,7093,5018,3679,2805,2124,1589,1150,794,509,289,130,34],
[10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10326,10271,9663,8649,7221,5442,4262,3380,2661,2067,1565,1143,795,512,288,131,33],
[26353,25879,25325,24692,23979,23191,22325,21372,20347,19242,18045,16766,15387,13922,12348,10666,8913,7565,6436,5456,4610,3839,3163,2557,2023,1557,1150,804,520,294,134,35],
[28792,27142,25531,23969,22457,20986,19568,18187,16865,15580,14339,13150,12002,10903,9849,8847,7884,6981,6128,5320,4564,3863,3218,2630,2093,1617,1197,841,554,318,143,37],
[39680,37898,36041,34110,32040,29946,27878,25843,23819,21819,19847,17895,15990,14282,12710,11247,9889,8624,7452,6382,5400,4502,3691,2970,2332,1761,1288,886,552,302,133,32],
[13050,12630,12306,12094,12050,12050,12050,12050,12050,12050,12050,12050,12015,11810,11492,11075,10572,9980,9303,8539,7692,6752,5705,4556,3280,2224,1527,1007,619,335,147,35],
[3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3580,3197,2177,1194,685,357,151,36],
[2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,2429,1402,728,370,150,37]],

[[1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,1780,974,436,170,37],
[2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2608,2508,1734,825,406,164,37],
[9174,8723,8402,8289,8289,8289,8289,8289,8289,8289,8289,8289,8289,8289,8289,8289,8289,8275,8041,7657,7148,6534,5816,4988,4043,2985,1865,1172,702,372,162,38],
[42209,40443,38591,36550,34452,32386,30346,28345,26376,24441,22542,20666,18820,16997,15208,13440,11706,10003,8542,7237,6075,5041,4116,3291,2568,1943,1406,965,608,328,139,33],
[28596,26969,25382,23861,22374,20948,19559,18216,16916,15658,14442,13265,12137,11038,10004,9002,8052,7140,6287,5472,4709,3997,3335,2738,2185,1687,1260,882,576,337,153,41],
[27236,26711,26095,25382,24581,23695,22714,21645,20490,19241,17899,16458,14912,13261,11480,9734,8401,7260,6256,5357,4549,3816,3168,2577,2046,1583,1180,827,540,309,140,37],
[11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11637,11466,10724,9599,8059,6203,4977,4034,3258,2603,2044,1562,1151,807,523,300,138,35],
[8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8271,8039,5824,4386,3427,2675,2071,1568,1151,807,522,300,136,36],
[8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8278,8045,5822,4382,3422,2669,2069,1566,1148,802,518,300,137,37],
[11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11630,11459,10720,9592,8058,6195,4968,4035,3261,2603,2041,1558,1152,805,525,298,137,34],
[27230,26705,26089,25375,24575,23685,22707,21638,20483,19235,17894,16457,14909,13252,11473,9728,8394,7249,6248,5349,4544,3812,3156,2570,2042,1580,1174,825,539,308,141,37],
[28618,26977,25400,23870,22397,20959,19573,18230,16926,15669,14447,13281,12139,11060,10012,9018,8059,7158,6292,5485,4719,4011,3348,2745,2192,1698,1266,887,577,339,155,39],
[42196,40447,38574,36559,34442,32376,30341,28335,26376,24437,22529,20656,18823,17001,15210,13440,11702,9992,8527,7234,6079,5040,4112,3281,2567,1944,1405,963,609,329,138,33],
[9177,8729,8407,8292,8292,8292,8292,8292,8292,8292,8292,8292,8292,8292,8292,8292,8292,8280,8048,7664,7153,6543,5815,4990,4049,2985,1864,1170,700,373,161,39],
[2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2604,2506,1732,828,408,164,39],
[1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,1777,965,432,173,39]],

[[1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,1196,598,206,44],
[1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1745,1309,514,196,45],
[6126,5634,5302,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5231,5050,4647,4102,3418,2594,1623,862,439,179,41],
[43562,41821,39958,37890,35783,33731,31732,29758,27840,25960,24113,22318,20565,18833,17146,15494,13873,12286,10732,9207,7701,6221,4943,3894,3015,2262,1630,1115,704,401,183,52],
[28599,26991,25441,23942,22496,21089,19730,18414,17131,15892,14688,13532,12409,11328,10286,9279,8330,7406,6528,5711,4922,4193,3514,2879,2309,1793,1332,933,609,337,141,31],
[28021,27439,26751,25956,25060,24067,22979,21786,20498,19103,17599,15992,14254,12396,10672,9340,8171,7130,6197,5349,4573,3864,3225,2640,2112,1641,1228,872,571,331,153,42],
[12957,12957,12957,12957,12957,12957,12957,12957,12957,12957,12957,12957,12957,12957,12665,11822,10592,8953,7026,5747,4749,3921,3210,2591,2053,1584,1181,836,548,317,146,38],
[9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9337,9042,6686,5146,4108,3287,2619,2054,1576,1169,821,538,313,143,37],
[9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9345,9052,6684,5148,4112,3288,2622,2055,1577,1170,823,539,314,147,40],
[12955,12955,12955,12955,12955,12955,12955,12955,12955,12955,12955,12955,12955,12955,12655,11809,10582,8943,7023,5741,4746,3919,3207,2593,2051,1583,1182,834,548,315,145,36],
[28014,27436,26748,25951,25059,24062,22971,21780,20491,19097,17597,15985,14250,12394,10679,9333,8176,7130,6198,5348,4571,3867,3213,2634,2108,1641,1226,869,570,332,151,40],
[28602,26986,25436,23938,22490,21089,19725,18409,17127,15891,14689,13533,12410,11329,10282,9278,8325,7409,6531,5704,4928,4197,3513,2883,2307,1792,1334,932,610,338,145,33],
[43571,41825,39965,37897,35790,33740,31732,29766,27839,25960,24123,22319,20563,18834,17143,15508,13880,12283,10729,9199,7692,6220,4940,3895,3010,2257,1629,1119,704,401,182,52],
[6123,5633,5302,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5233,5055,4653,4101,3415,2594,1625,864,439,179,42],
[1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1742,1303,517,194,45],
[1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,1192,591,200,45]],

[[701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,701,301,57],
[1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,1012,885,261,53],
[3816,3270,2909,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2843,2837,2547,2044,1368,598,232,54],
[43650,41929,40078,37993,35894,33856,31871,29931,28052,26223,24444,22710,21035,19402,17815,16276,14777,13318,11904,10532,9189,7877,6601,5354,4137,2948,2044,1362,838,453,191,46],
[28902,27311,25791,24320,22900,21518,20178,18880,17620,16392,15205,14049,12928,11845,10789,9784,8814,7875,6980,6125,5316,4551,3832,3163,2555,1996,1496,1058,696,401,186,48],
[28725,28083,27318,26439,25449,24345,23133,21817,20382,18830,17161,15368,13418,11740,10380,9191,8123,7147,6256,5433,4675,3978,3338,2756,2222,1745,1317,940,627,372,171,45],
[14309,14309,14309,14309,14309,14309,14309,14309,14309,14309,14309,14309,14309,13889,12965,11647,9909,7924,6599,5553,4670,3903,3230,2634,2110,1640,1234,882,587,344,164,43],
[10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10466,10105,7613,5978,4866,3976,3243,2618,2080,1613,1206,861,571,334,155,42],
[10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10473,10112,7619,5981,4872,3979,3251,2623,2084,1616,1208,865,575,340,160,43],
[14312,14312,14312,14312,14312,14312,14312,14312,14312,14312,14312,14312,14312,13893,12967,11652,9908,7927,6606,5555,4668,3903,3229,2637,2105,1641,1234,881,585,346,160,44],
[28725,28082,27319,26439,25445,24339,23127,21806,20374,18825,17159,15357,13413,11737,10378,9191,8113,7146,6253,5427,4673,3975,3337,2750,2221,1740,1312,940,623,364,171,45],
[28903,27317,25796,24324,22900,21520,20186,18887,17624,16395,15203,14049,12923,11847,10798,9785,8808,7875,6982,6130,5310,4554,3835,3164,2554,1993,1498,1065,698,403,189,48],
[43641,41915,40074,37986,35889,33847,31865,29925,28047,26214,24439,22709,21028,19395,17810,16271,14771,13317,11899,10523,9187,7876,6606,5359,4132,2935,2036,1360,841,455,191,47],
[3814,3270,2910,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2847,2838,2550,2052,1361,595,226,49],
[1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,1011,883,265,54],
[694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,694,292,51]],

[[300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,300,89],
[434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,434,79],
[2228,1624,1220,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,1141,936,412,72],
[42292,40594,38764,36676,34581,32542,30573,28671,26818,25041,23317,21655,20057,18514,17031,15594,14220,12902,11635,10416,9248,8126,7048,6014,5016,4056,3127,2220,1334,666,283,72],
[29551,27992,26493,25049,23653,22293,20977,19695,18444,17225,16034,14884,13760,12659,11597,10571,9568,8607,7677,6783,5927,5112,4336,3617,2943,2320,1756,1259,832,483,217,52],
[29431,28713,27867,26889,25799,24578,23237,21776,20192,18475,16619,14617,12950,11564,10351,9246,8228,7297,6428,5620,4871,4173,3531,2935,2388,1890,1447,1046,704,418,196,53],
[15701,15701,15701,15701,15701,15701,15701,15701,15701,15701,15701,15701,15179,14187,12788,10958,8918,7554,6450,5514,4692,3968,3317,2731,2207,1742,1325,961,649,389,188,54],
[11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11671,11256,8629,6906,5719,4761,3971,3283,2685,2161,1689,1285,930,627,377,186,55],
[11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11667,11251,8620,6896,5710,4752,3965,3275,2679,2151,1685,1277,924,623,374,181,47],
[15708,15708,15708,15708,15708,15708,15708,15708,15708,15708,15708,15708,15192,14198,12800,10967,8925,7552,6458,5520,4700,3972,3323,2737,2214,1747,1335,972,657,397,192,54],
[29436,28726,27872,26898,25806,24586,23245,21784,20201,18482,16626,14615,12952,11569,10353,9249,8234,7299,6431,5624,4877,4180,3532,2935,2385,1885,1435,1040,699,413,197,55],
[29541,27981,26486,25045,23641,22285,20963,19677,18431,17218,16027,14871,13750,12657,11599,10566,9564,8601,7674,6775,5920,5104,4339,3611,2936,2320,1755,1255,830,480,216,50],
[42303,40602,38774,36689,34591,32549,30584,28676,26834,25047,23330,21659,20061,18526,17038,15601,14230,12911,11641,10425,9258,8135,7059,6025,5026,4065,3138,2234,1350,672,289,76],
[2218,1611,1210,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,1133,933,409,71],
[431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,431,82],
[297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,297,86]],

[[38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38],
[56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56,56],
[1379,696,239,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130,130],
[38753,37075,35276,33190,31082,29047,27086,25209,23398,21654,19995,18396,16877,15423,14038,12727,11478,10304,9194,8143,7169,6257,5408,4621,3888,3219,2606,2042,1525,1049,601,174],
[31045,29506,28032,26612,25234,23892,22586,21325,20068,18859,17674,16508,15377,14266,13178,12121,11083,10075,9094,8142,7219,6327,5468,4637,3858,3118,2429,1802,1238,747,364,100],
[30250,29455,28520,27442,26230,24890,23415,21806,20045,18145,16080,14396,12998,11744,10603,9553,8575,7663,6804,6005,5258,4555,3896,3279,2709,2183,1695,1263,876,547,277,85],
[17178,17178,17178,17178,17178,17178,17178,17178,17178,17178,17170,16597,15543,14079,12161,10062,8645,7489,6508,5633,4854,4148,3510,2928,2399,1919,1490,1106,769,478,249,77],
[12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12958,12484,9746,7933,6677,5654,4801,4060,3405,2822,2295,1831,1413,1044,725,455,236,74],
[12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12962,12486,9741,7926,6669,5648,4801,4054,3399,2815,2288,1824,1409,1042,720,448,227,71],
[17171,17171,17171,17171,17171,17171,17171,17171,17171,17171,17163,16592,15534,14065,12153,10053,8637,7492,6502,5631,4848,4142,3502,2919,2391,1913,1485,1101,763,473,243,75],
[30248,29455,28516,27441,26229,24890,23412,21806,20043,18138,16083,14399,13008,11751,10609,9559,8577,7667,6811,6012,5256,4557,3900,3284,2712,2182,1702,1267,882,552,281,85],
[31048,29519,28033,26611,25240,23910,22593,21311,20072,18866,17680,16508,15378,14272,13180,12121,11086,10075,9095,8143,7214,6324,5465,4643,3863,3124,2433,1805,1242,753,365,102],
[38753,37069,35267,33189,31082,29047,27089,25209,23393,21655,19990,18396,16872,15414,14041,12728,11478,10307,9196,8147,7169,6259,5410,4622,3888,3218,2607,2044,1524,1046,601,175],
[1382,709,242,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132,132],
[55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55,55],
[35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35,35]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1449,697,162,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[31917,30263,28494,26433,24311,22269,20318,18453,16666,14974,13364,11836,10389,9034,7757,6559,5461,4441,3518,2673,1934,1286,744,319,43,0,0,0,0,0,0,0],
[33629,32117,30665,29260,27909,26580,25279,24017,22775,21564,20367,19196,18042,16904,15787,14683,13594,12525,11470,10422,9387,8359,7335,6301,5235,4055,2951,1981,1161,503,65,0],
[31417,30538,29502,28318,26987,25506,23881,22100,20149,18056,16358,14914,13633,12453,11363,10341,9390,8491,7638,6828,6064,5341,4658,4009,3402,2822,2286,1784,1311,884,469,11],
[18828,18828,18828,18828,18828,18828,18828,18828,18828,18813,18202,17104,15585,13584,11434,9960,8765,7730,6807,5981,5225,4528,3894,3306,2769,2271,1818,1406,1031,693,394,112],
[14447,14447,14447,14447,14447,14447,14447,14447,14447,14447,14447,14447,14447,14447,13925,11071,9168,7849,6763,5855,5055,4344,3706,3122,2596,2119,1684,1295,947,640,368,127],
[14443,14443,14443,14443,14443,14443,14443,14443,14443,14443,14443,14443,14443,14443,13926,11061,9162,7839,6753,5844,5044,4334,3696,3110,2586,2109,1674,1284,935,627,362,124],
[18841,18841,18841,18841,18841,18841,18841,18841,18841,18828,18214,17113,15597,13597,11444,9971,8772,7744,6824,5996,5234,4541,3903,3317,2777,2279,1821,1407,1032,697,401,114],
[31407,30529,29490,28306,26977,25498,23867,22096,20145,18048,16336,14903,13620,12440,11355,10336,9382,8476,7622,6818,6057,5333,4649,4006,3395,2819,2289,1785,1323,884,469,9],
[33620,32105,30666,29261,27899,26569,25279,24016,22770,21556,20360,19192,18029,16897,15774,14680,13592,12519,11459,10414,9377,8350,7329,6297,5228,4052,2947,1980,1154,499,63,0],
[31919,30277,28498,26441,24315,22280,20322,18456,16674,14982,13365,11840,10392,9034,7761,6571,5466,4447,3520,2682,1935,1291,747,323,46,0,0,0,0,0,0,0],
[1446,690,161,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1687,844,234,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28162,26542,24797,22776,20627,18581,16619,14769,13015,11355,9792,8328,6967,5707,4547,3495,2551,1720,1023,466,83,0,0,0,0,0,0,0,0,0,0,0],
[33246,31753,30326,28939,27597,26286,25000,23738,22503,21286,20081,18888,17709,16540,15379,14219,13056,11883,10705,9500,8220,6770,5354,4069,2913,1905,1052,389,7,0,0,0],
[32269,31297,30154,28846,27375,25746,23944,21984,19837,18075,16604,15283,14067,12939,11884,10884,9942,9036,8182,7361,6579,5822,5102,4408,3743,3093,2459,1822,1128,313,0,0],
[20429,20429,20429,20429,20429,20429,20429,20429,20421,19808,18672,17099,15037,12820,11314,10062,8975,8012,7132,6322,5574,4883,4239,3635,3070,2543,2044,1583,1138,693,94,0],
[15929,15929,15929,15929,15929,15929,15929,15929,15929,15929,15929,15929,15929,15374,12401,10420,9033,7889,6924,6068,5306,4610,3969,3389,2848,2351,1889,1466,1069,687,273,0],
[15906,15906,15906,15906,15906,15906,15906,15906,15906,15906,15906,15906,15906,15356,12381,10395,9012,7869,6906,6051,5288,4592,3952,3370,2831,2335,1874,1449,1054,680,271,0],
[20449,20449,20449,20449,20449,20449,20449,20449,20440,19822,18696,17126,15057,12841,11326,10074,8985,8016,7137,6327,5580,4884,4242,3641,3080,2547,2050,1587,1144,695,94,0],
[32268,31294,30152,28849,27375,25739,23946,21980,19824,18077,16596,15281,14068,12941,11884,10889,9944,9042,8182,7359,6580,5826,5103,4413,3746,3095,2461,1830,1131,316,0,0],
[33246,31758,30327,28941,27602,26287,25003,23744,22509,21281,20078,18893,17713,16540,15376,14213,13054,11886,10709,9497,8220,6769,5356,4067,2909,1902,1053,393,7,0,0,0],
[28163,26545,24796,22777,20627,18575,16618,14768,13014,11358,9797,8336,6969,5708,4548,3496,2549,1724,1027,465,84,0,0,0,0,0,0,0,0,0,0,0],
[1689,847,236,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1978,1035,333,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[25553,23968,22245,20292,18109,16040,14075,12232,10498,8875,7369,5973,4695,3537,2514,1621,880,314,1,0,0,0,0,0,0,0,0,0,0,0,0,0],
[32222,30749,29342,27971,26640,25336,24055,22797,21550,20322,19101,17889,16666,15449,14224,12982,11705,10374,8913,7261,5743,4362,3123,2033,1122,412,4,0,0,0,0,0],
[32742,31663,30402,28960,27355,25550,23552,21370,19557,18039,16684,15441,14274,13182,12141,11148,10204,9295,8420,7576,6752,5953,5174,4403,3628,2838,1939,872,118,0,0,0],
[21915,21915,21915,21915,21915,21915,21915,21915,21320,20171,18558,16426,14143,12583,11291,10159,9142,8210,7358,6558,5813,5112,4450,3819,3220,2642,2085,1523,912,56,0,0],
[17347,17347,17347,17347,17347,17347,17347,17347,17347,17347,17347,17347,16768,13689,11626,10176,8983,7964,7059,6241,5495,4796,4161,3559,2998,2471,1974,1490,999,404,0,0],
[17345,17345,17345,17345,17345,17345,17345,17345,17345,17345,17345,17345,16764,13676,11621,10166,8970,7956,7047,6229,5485,4789,4153,3554,2993,2465,1962,1475,992,401,0,0],
[21902,21902,21902,21902,21902,21902,21902,21902,21307,20165,18548,16421,14141,12577,11285,10155,9139,8216,7358,6563,5817,5112,4447,3818,3220,2644,2079,1523,910,57,0,0],
[32753,31671,30410,28974,27354,25553,23575,21376,19568,18049,16687,15449,14279,13184,12146,11155,10207,9297,8422,7573,6752,5958,5172,4402,3633,2838,1942,879,127,0,0,0],
[32222,30761,29352,27984,26645,25346,24065,22806,21562,20329,19110,17888,16674,15455,14229,12987,11705,10375,8914,7266,5748,4360,3121,2038,1122,413,5,0,0,0,0,0],
[25551,23964,22250,20285,18105,16022,14069,12231,10491,8872,7356,5968,4691,3531,2504,1616,877,312,2,0,0,0,0,0,0,0,0,0,0,0,0,0],
[1971,1036,330,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2310,1257,457,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23624,22057,20369,18470,16259,14156,12190,10349,8624,7041,5583,4256,3067,2037,1164,478,44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30972,29524,28134,26781,25456,24164,22888,21625,20372,19131,17880,16633,15373,14093,12777,11407,9914,8153,6500,4988,3625,2418,1395,579,53,0,0,0,0,0,0,0],
[32910,31714,30318,28722,26935,24953,22734,20821,19267,17872,16588,15387,14255,13174,12147,11155,10202,9276,8369,7482,6604,5740,4866,3953,2950,1692,644,8,0,0,0,0],
[23263,23263,23263,23263,23263,23263,23263,22736,21580,19939,17757,15407,13792,12452,11273,10215,9238,8338,7490,6690,5934,5210,4518,3845,3190,2533,1852,1028,23,0,0,0],
[18719,18719,18719,18719,18719,18719,18719,18719,18719,18719,18719,18122,14941,12807,11299,10047,8980,8021,7154,6359,5610,4919,4261,3638,3041,2470,1898,1297,520,0,0,0],
[18722,18722,18722,18722,18722,18722,18722,18722,18722,18722,18722,18123,14938,12806,11297,10043,8977,8020,7152,6352,5608,4915,4257,3634,3037,2462,1890,1293,517,0,0,0],
[23263,23263,23263,23263,23263,23263,23263,22741,21582,19940,17760,15407,13789,12450,11274,10214,9242,8336,7491,6695,5932,5213,4519,3847,3189,2528,1851,1026,22,0,0,0],
[32895,31698,30309,28716,26921,24934,22732,20806,19247,17856,16570,15376,14245,13167,12141,11148,10195,9265,8364,7478,6597,5730,4856,3948,2948,1698,644,7,0,0,0,0],
[30984,29529,28142,26780,25460,24170,22891,21633,20377,19136,17885,16638,15377,14093,12781,11405,9908,8162,6498,4991,3625,2423,1394,581,56,0,0,0,0,0,0,0],
[23623,22056,20372,18474,16260,14164,12189,10347,8626,7048,5589,4258,3074,2040,1165,483,46,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2309,1258,458,6,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2713,1535,619,53,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22112,20572,18921,17065,14848,12717,10725,8880,7173,5621,4219,2968,1887,994,325,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29669,28241,26865,25520,24208,22922,21644,20376,19109,17849,16580,15288,13963,12593,11142,9488,7655,5968,4437,3080,1905,934,231,0,0,0,0,0,0,0,0,0],
[32846,31512,29963,28202,26219,24005,21970,20348,18911,17586,16354,15188,14077,13004,11972,10971,9986,9029,8078,7131,6173,5187,4129,2796,1474,453,0,0,0,0,0,0],
[24476,24476,24476,24476,24476,24476,24046,22912,21253,19030,16600,14933,13550,12327,11223,10206,9260,8361,7513,6704,5922,5166,4427,3686,2939,2125,1035,1,0,0,0,0],
[20037,20037,20037,20037,20037,20037,20037,20037,20037,20037,19434,16156,13952,12387,11086,9970,8965,8052,7205,6405,5660,4947,4263,3604,2958,2302,1588,616,0,0,0,0],
[20041,20041,20041,20041,20041,20041,20041,20041,20041,20041,19438,16157,13951,12387,11081,9972,8964,8049,7203,6405,5657,4948,4265,3603,2952,2294,1590,611,0,0,0,0],
[24484,24484,24484,24484,24484,24484,24055,22917,21257,19026,16605,14930,13544,12322,11218,10203,9253,8359,7512,6702,5920,5165,4424,3686,2939,2128,1029,1,0,0,0,0],
[32826,31499,29949,28188,26205,23998,21955,20340,18904,17583,16348,15182,14076,13004,11974,10972,9993,9035,8080,7137,6178,5192,4125,2796,1478,446,0,0,0,0,0,0],
[29670,28250,26871,25524,24216,22916,21644,20377,19118,17856,16582,15294,13969,12598,11142,9483,7650,5963,4430,3070,1897,933,228,0,0,0,0,0,0,0,0,0],
[22105,20565,18908,17063,14841,12713,10726,8877,7176,5614,4202,2955,1877,985,319,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2717,1538,624,52,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3182,1862,827,135,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20905,19382,17756,15952,13751,11577,9559,7705,6009,4489,3132,1963,1007,298,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28415,27002,25635,24308,23005,21716,20440,19161,17884,16599,15281,13939,12533,11035,9283,7384,5644,4093,2716,1548,627,41,0,0,0,0,0,0,0,0,0,0],
[32498,31015,29302,27342,25139,22967,21284,19805,18444,17177,15982,14835,13730,12667,11622,10598,9587,8574,7561,6521,5418,4181,2633,1293,296,0,0,0,0,0,0,0],
[25594,25594,25594,25594,25594,25282,24177,22511,20243,17740,16012,14583,13323,12176,11117,10126,9186,8287,7427,6593,5775,4961,4147,3295,2360,1022,0,0,0,0,0,0],
[21307,21307,21307,21307,21307,21307,21307,21307,21307,20716,17346,15074,13453,12107,10945,9895,8931,8041,7194,6394,5628,4886,4160,3439,2698,1878,686,0,0,0,0,0],
[21301,21301,21301,21301,21301,21301,21301,21301,21301,20707,17329,15058,13443,12091,10932,9879,8920,8023,7177,6378,5614,4875,4148,3425,2684,1873,688,0,0,0,0,0],
[25593,25593,25593,25593,25593,25283,24180,22515,20251,17742,16014,14589,13325,12180,11123,10129,9193,8297,7432,6601,5776,4968,4153,3306,2362,1022,0,0,0,0,0,0],
[32507,31022,29309,27349,25142,22974,21284,19804,18448,17187,15985,14842,13736,12664,11618,10592,9580,8573,7562,6519,5424,4182,2633,1289,297,0,0,0,0,0,0,0],
[28416,27002,25633,24311,22999,21715,20441,19158,17875,16586,15274,13934,12528,11026,9279,7384,5652,4093,2716,1546,624,47,0,0,0,0,0,0,0,0,0,0],
[20901,19380,17748,15951,13747,11572,9565,7709,6010,4486,3133,1973,1011,298,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3182,1871,827,133,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3744,2261,1082,248,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19917,18423,16810,15058,12902,10679,8622,6745,5065,3564,2271,1199,392,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[27213,25818,24467,23144,21843,20556,19274,17987,16692,15374,14015,12596,11085,9317,7333,5535,3920,2511,1329,435,0,0,0,0,0,0,0,0,0,0,0,0],
[31990,30334,28414,26224,23926,22164,20631,19247,17935,16703,15526,14390,13285,12208,11137,10082,9018,7938,6816,5610,4159,2513,1150,180,0,0,0,0,0,0,0,0],
[26572,26572,26572,26572,26391,25352,23697,21406,18810,17023,15553,14253,13067,11968,10937,9952,9014,8102,7215,6335,5464,4573,3631,2550,1014,0,0,0,0,0,0,0],
[22525,22525,22525,22525,22525,22525,22525,22525,21950,18499,16158,14491,13094,11890,10795,9791,8852,7957,7112,6293,5492,4703,3913,3089,2161,759,0,0,0,0,0,0],
[22522,22522,22522,22522,22522,22522,22522,22522,21946,18486,16148,14483,13081,11882,10787,9781,8841,7946,7103,6283,5483,4693,3898,3070,2150,756,0,0,0,0,0,0],
[26575,26575,26575,26575,26393,25354,23699,21405,18813,17033,15554,14255,13071,11973,10938,9956,9014,8104,7220,6340,5469,4579,3638,2559,1015,0,0,0,0,0,0,0],
[31980,30327,28410,26225,23929,22164,20633,19245,17932,16701,15522,14395,13285,12206,11140,10078,9013,7937,6820,5616,4163,2515,1151,180,0,0,0,0,0,0,0,0],
[27198,25804,24454,23133,21833,20547,19266,17981,16678,15362,14010,12588,11076,9317,7341,5533,3924,2505,1330,431,0,0,0,0,0,0,0,0,0,0,0,0],
[19928,18424,16824,15065,12906,10678,8633,6757,5072,3568,2275,1204,397,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[3740,2272,1084,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4411,2736,1398,412,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19119,17647,16076,14352,12270,9974,7872,5982,4291,2819,1581,624,37,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26079,24701,23358,22044,20748,19458,18172,16870,15551,14201,12788,11280,9585,7521,5624,3927,2449,1230,331,0,0,0,0,0,0,0,0,0,0,0,0,0],
[31263,29410,27254,24852,22981,21395,19954,18620,17355,16148,14981,13845,12726,11630,10525,9414,8283,7095,5795,4166,2456,1056,104,0,0,0,0,0,0,0,0,0],
[27480,27480,27480,27419,26486,24856,22547,19859,18014,16496,15153,13933,12795,11725,10702,9723,8767,7828,6895,5956,4988,3957,2743,1022,0,0,0,0,0,0,0,0],
[23660,23660,23660,23660,23660,23660,23660,23123,19583,17181,15469,14033,12785,11652,10607,9625,8685,7790,6919,6067,5217,4361,3454,2424,827,0,0,0,0,0,0,0],
[23662,23662,23662,23662,23662,23662,23662,23123,19577,17179,15465,14028,12782,11646,10601,9617,8680,7786,6917,6063,5212,4351,3443,2420,826,0,0,0,0,0,0,0],
[27475,27475,27475,27415,26479,24851,22541,19853,18008,16491,15148,13928,12792,11720,10697,9714,8760,7820,6891,5953,4984,3955,2742,1025,0,0,0,0,0,0,0,0],
[31271,29415,27266,24867,22989,21399,19959,18621,17360,16152,14982,13852,12736,11632,10530,9422,8284,7095,5796,4168,2453,1058,104,0,0,0,0,0,0,0,0,0],
[26080,24697,23358,22042,20748,19459,18173,16870,15555,14197,12781,11276,9582,7514,5613,3925,2452,1232,337,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19116,17641,16069,14346,12258,9962,7868,5974,4284,2813,1581,623,34,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4404,2743,1397,421,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5195,3316,1786,633,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18468,17008,15458,13771,11796,9423,7258,5333,3637,2191,1022,211,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[25021,23652,22320,21017,19726,18434,17138,15826,14476,13073,11591,9962,7931,5902,4090,2527,1238,305,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30377,28283,25849,23791,22130,20642,19268,17974,16733,15538,14374,13229,12096,10956,9809,8625,7377,6007,4245,2454,995,55,0,0,0,0,0,0,0,0,0,0],
[28264,28264,28264,27516,25945,23629,20845,18945,17380,16001,14743,13573,12465,11404,10383,9380,8398,7410,6415,5383,4258,2905,1039,0,0,0,0,0,0,0,0,0],
[24758,24758,24758,24758,24758,24758,24266,20647,18187,16431,14952,13672,12498,11412,10390,9406,8469,7548,6643,5734,4811,3824,2692,896,0,0,0,0,0,0,0,0],
[24749,24749,24749,24749,24749,24749,24255,20634,18176,16422,14941,13661,12487,11403,10378,9397,8455,7543,6634,5724,4799,3813,2688,896,0,0,0,0,0,0,0,0],
[28262,28262,28262,27517,25940,23632,20845,18939,17372,15992,14734,13563,12458,11395,10375,9373,8394,7402,6409,5379,4254,2905,1037,0,0,0,0,0,0,0,0,0],
[30374,28284,25854,23787,22123,20646,19272,17980,16749,15552,14385,13241,12104,10965,9815,8632,7388,6016,4251,2457,1000,57,0,0,0,0,0,0,0,0,0,0],
[25024,23654,22319,21018,19727,18432,17133,15814,14468,13077,11591,9961,7931,5902,4094,2529,1239,304,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18459,16999,15459,13771,11793,9418,7262,5331,3636,2190,1024,215,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5204,3329,1788,637,5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6159,4024,2266,925,81,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17923,16487,14966,13310,11407,9005,6768,4797,3089,1669,592,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24055,22695,21373,20069,18778,17482,16179,14840,13459,12011,10435,8607,6407,4447,2753,1356,340,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29286,26905,24613,22865,21331,19915,18590,17325,16101,14908,13736,12577,11404,10219,8998,7704,6271,4429,2533,999,34,0,0,0,0,0,0,0,0,0,0,0],
[28965,28965,28475,26972,24675,21787,19815,18205,16792,15502,14296,13155,12061,11001,9963,8934,7902,6852,5755,4553,3080,1072,0,0,0,0,0,0,0,0,0,0],
[25771,25771,25771,25771,25771,25341,21651,19139,17346,15829,14512,13301,12180,11120,10094,9114,8151,7190,6228,5239,4178,2948,970,0,0,0,0,0,0,0,0,0],
[25769,25769,25769,25769,25769,25336,21640,19126,17330,15817,14503,13293,12171,11111,10088,9106,8142,7185,6220,5228,4164,2942,970,0,0,0,0,0,0,0,0,0],
[28971,28971,28481,26977,24683,21800,19828,18220,16804,15511,14304,13162,12066,11004,9964,8933,7903,6852,5755,4556,3081,1067,0,0,0,0,0,0,0,0,0,0],
[29284,26900,24594,22861,21318,19901,18581,17315,16094,14906,13734,12562,11400,10218,8993,7703,6278,4426,2520,995,32,0,0,0,0,0,0,0,0,0,0,0],
[24043,22687,21367,20066,18774,17477,16167,14830,13455,12002,10437,8602,6406,4449,2754,1358,339,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17925,16481,14958,13310,11413,9005,6776,4799,3088,1670,589,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6167,4027,2273,924,82,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7345,4894,2872,1297,238,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17479,16053,14552,12930,11084,8716,6396,4357,2631,1242,263,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23158,21814,20501,19201,17903,16602,15276,13922,12502,10991,9311,7166,5019,3150,1605,455,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[27947,25449,23588,21982,20530,19171,17878,16628,15415,14225,13032,11840,10628,9380,8056,6592,4702,2677,1039,26,0,0,0,0,0,0,0,0,0,0,0,0],
[29599,29358,27981,25718,22730,20682,19029,17577,16250,15010,13835,12708,11614,10532,9462,8388,7290,6130,4859,3274,1123,0,0,0,0,0,0,0,0,0,0,0],
[26720,26720,26720,26720,26357,22603,20038,18206,16659,15310,14069,12913,11820,10755,9735,8725,7721,6704,5655,4522,3197,1043,0,0,0,0,0,0,0,0,0,0],
[26731,26731,26731,26731,26365,22605,20042,18213,16659,15312,14069,12916,11817,10752,9728,8722,7717,6699,5648,4512,3193,1045,0,0,0,0,0,0,0,0,0,0],
[29586,29342,27964,25707,22719,20672,19018,17564,16242,14999,13824,12700,11612,10534,9470,8391,7289,6130,4859,3278,1122,0,0,0,0,0,0,0,0,0,0,0],
[27955,25458,23590,21993,20538,19180,17884,16640,15426,14229,13043,11847,10638,9379,8057,6591,4699,2677,1039,25,0,0,0,0,0,0,0,0,0,0,0,0],
[23141,21801,20486,19188,17897,16593,15267,13913,12496,10986,9302,7157,5010,3141,1599,448,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17481,16056,14552,12932,11097,8719,6395,4356,2629,1241,264,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7346,4893,2877,1294,235,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8774,5965,3628,1782,468,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17112,15712,14227,12626,10843,8543,6102,3997,2226,870,42,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22332,21005,19699,18396,17103,15793,14453,13071,11610,10024,8180,5835,3750,2002,668,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26355,24317,22643,21144,19754,18423,17155,15923,14705,13499,12288,11056,9783,8439,6954,5105,2916,1147,35,0,0,0,0,0,0,0,0,0,0,0,0,0],
[30119,28929,26724,23647,21505,19805,18319,16962,15697,14491,13332,12199,11088,9985,8865,7717,6506,5166,3498,1195,0,0,0,0,0,0,0,0,0,0,0,0],
[27617,27617,27617,27327,23516,20902,19040,17460,16079,14810,13624,12497,11397,10338,9290,8240,7171,6063,4860,3445,1126,0,0,0,0,0,0,0,0,0,0,0],
[27608,27608,27608,27319,23502,20889,19027,17446,16069,14795,13612,12481,11385,10328,9280,8233,7164,6056,4850,3439,1125,0,0,0,0,0,0,0,0,0,0,0],
[30112,28931,26725,23635,21503,19802,18313,16958,15693,14493,13334,12205,11095,9986,8860,7717,6506,5168,3500,1201,0,0,0,0,0,0,0,0,0,0,0,0],
[26361,24322,22647,21144,19749,18432,17158,15925,14710,13500,12284,11057,9784,8438,6961,5105,2915,1146,37,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22337,21005,19696,18400,17106,15794,14454,13073,11612,10029,8179,5836,3750,2001,664,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17118,15716,14229,12636,10849,8544,6112,3990,2236,876,47,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[8775,5976,3630,1776,468,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[45,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[884,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9827,7343,4594,2409,802,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16823,15435,13974,12403,10662,8503,5917,3707,1901,572,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21586,20266,18967,17667,16365,15043,13688,12276,10770,9093,6969,4606,2600,1028,50,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[25110,23336,21774,20342,18996,17703,16445,15213,13982,12754,11510,10221,8869,7382,5611,3267,1330,67,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29802,27693,24517,22286,20536,19013,17628,16335,15105,13918,12758,11617,10476,9319,8132,6875,5481,3752,1303,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28445,28445,28240,24382,21725,19829,18223,16819,15521,14304,13148,12018,10925,9836,8742,7623,6462,5193,3691,1213,0,0,0,0,0,0,0,0,0,0,0,0],
[28437,28437,28233,24368,21708,19815,18209,16805,15507,14295,13134,12010,10913,9825,8733,7619,6451,5180,3687,1215,0,0,0,0,0,0,0,0,0,0,0,0],
[29804,27695,24515,22286,20537,19016,17628,16335,15103,13917,12759,11616,10479,9327,8134,6875,5483,3757,1299,0,0,0,0,0,0,0,0,0,0,0,0,0],
[25116,23344,21786,20355,19001,17711,16451,15220,13995,12766,11519,10236,8874,7388,5615,3280,1331,71,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21588,20262,18959,17668,16362,15039,13684,12271,10766,9085,6963,4601,2593,1018,50,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16815,15428,13963,12400,10657,8510,5915,3704,1898,569,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9840,7351,4594,2410,809,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[886,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2574,470,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10676,8675,5849,3233,1277,94,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16591,15216,13776,12229,10527,8509,5824,3494,1628,331,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20901,19584,18282,16987,15684,14350,12971,11525,9956,8175,5791,3461,1568,261,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24065,22433,20961,19577,18259,16981,15724,14487,13250,11994,10690,9334,7851,6134,3764,1600,139,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28654,25409,23063,21260,19702,18287,16969,15713,14501,13317,12144,10968,9792,8559,7261,5823,4049,1456,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29187,29068,25176,22483,20564,18934,17504,16182,14944,13760,12600,11477,10351,9216,8055,6841,5508,3923,1309,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29196,29077,25178,22482,20566,18935,17511,16186,14944,13758,12598,11471,10349,9218,8050,6834,5499,3919,1308,0,0,0,0,0,0,0,0,0,0,0,0,0],
[28644,25401,23049,21252,19692,18275,16953,15702,14490,13306,12136,10966,9781,8554,7254,5816,4047,1454,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24072,22439,20967,19585,18264,16990,15738,14497,13252,11997,10695,9339,7855,6139,3770,1603,141,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20893,19574,18275,16976,15674,14345,12968,11519,9954,8169,5790,3453,1571,265,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16593,15220,13773,12234,10534,8506,5827,3490,1624,334,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10680,8672,5847,3232,1276,91,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[2572,471,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[42,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5478,1973,119,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11420,9628,7408,4340,1928,343,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16420,15053,13635,12114,10448,8504,5836,3339,1385,142,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20286,18971,17673,16373,15061,13712,12317,10825,9198,7274,4678,2395,694,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23121,21589,20171,18828,17538,16272,15022,13767,12505,11201,9839,8373,6709,4455,2005,273,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26287,23782,21934,20340,18898,17557,16276,15037,13833,12641,11444,10232,8972,7638,6163,4368,1645,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29834,25939,23203,21263,19612,18166,16824,15564,14353,13162,12011,10849,9678,8477,7213,5823,4165,1414,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[29826,25921,23190,21252,19598,18156,16813,15552,14340,13153,11998,10841,9672,8468,7202,5812,4159,1414,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26286,23785,21932,20340,18900,17558,16278,15043,13836,12638,11443,10229,8972,7640,6166,4366,1646,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23134,21605,20185,18839,17543,16276,15024,13771,12503,11199,9836,8375,6706,4461,2006,276,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20280,18972,17673,16376,15059,13714,12320,10829,9201,7283,4672,2400,696,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16407,15047,13621,12107,10438,8499,5836,3334,1388,145,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11420,9626,7404,4342,1931,347,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[5482,1975,122,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[43,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[4129,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7693,4826,1360,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12095,10436,8517,5863,2855,771,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16294,14938,13536,12041,10409,8539,5972,3255,1191,20,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19699,18392,17099,15803,14479,13118,11696,10175,8471,6362,3600,1417,71,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22262,20803,19428,18110,16832,15562,14303,13032,11730,10369,8924,7309,5319,2574,507,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24530,22615,20989,19521,18154,16854,15600,14370,13153,11933,10694,9409,8041,6535,4723,1908,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26618,23857,21895,20229,18765,17404,16125,14894,13679,12500,11311,10108,8868,7560,6118,4392,1536,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[26612,23850,21893,20221,18764,17402,16121,14890,13676,12494,11306,10104,8861,7554,6106,4387,1535,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24521,22610,20983,19507,18144,16844,15587,14361,13147,11931,10688,9403,8040,6535,4721,1908,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22255,20802,19432,18115,16832,15569,14306,13036,11729,10373,8924,7306,5320,2570,506,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19704,18403,17103,15805,14489,13128,11699,10176,8470,6362,3608,1425,72,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16294,14933,13534,12038,10399,8531,5966,3255,1197,19,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12091,10430,8510,5861,2851,770,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[7703,4829,1363,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[4123,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[6983,4024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9317,7239,4085,761,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12717,11157,9414,7316,4174,1444,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16222,14876,13485,12011,10410,8594,6268,3254,1033,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19179,17876,16588,15281,13951,12579,11124,9553,7762,5376,2580,574,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21459,20049,18710,17413,16142,14872,13594,12289,10941,9510,7939,6092,3390,877,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23247,21583,20090,18703,17381,16110,14864,13631,12389,11131,9818,8437,6910,5094,2257,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24474,22500,20818,19342,17969,16674,15422,14186,12982,11765,10532,9252,7901,6409,4618,1677,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[24459,22489,20802,19333,17957,16658,15407,14177,12972,11758,10525,9247,7894,6397,4613,1674,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23258,21592,20098,18708,17390,16119,14874,13637,12391,11135,9824,8436,6917,5097,2259,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21455,20051,18713,17415,16140,14865,13593,12289,10940,9504,7937,6091,3384,876,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19179,17880,16584,15281,13952,12580,11125,9556,7766,5373,2579,569,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16221,14881,13488,12014,10411,8595,6265,3248,1033,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12716,11155,9413,7314,4175,1439,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9315,7240,4080,755,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[6980,4022,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[9006,6897,3910,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10673,8892,6738,3265,223,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13325,11838,10210,8349,5971,2500,282,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16179,14849,13470,12017,10448,8680,6517,3361,911,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18712,17417,16126,14814,13473,12082,10600,8975,7086,4392,1595,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20689,19325,18012,16728,15450,14170,12868,11530,10113,8585,6834,4534,1465,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22171,20652,19244,17911,16625,15365,14111,12852,11579,10254,8853,7312,5500,2731,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23053,21359,19876,18485,17176,15910,14655,13429,12189,10924,9614,8229,6688,4839,1845,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[23047,21353,19869,18475,17165,15897,14645,13419,12182,10919,9606,8218,6680,4836,1845,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[22161,20640,19238,17907,16621,15356,14109,12857,11579,10258,8854,7311,5498,2735,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20703,19339,18022,16732,15459,14181,12877,11532,10114,8581,6836,4535,1466,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18709,17409,16119,14808,13468,12072,10593,8973,7084,4390,1595,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16176,14846,13468,12016,10441,8683,6515,3358,911,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13331,11846,10217,8352,5977,2503,287,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10669,8892,6739,3265,223,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[9011,6898,3906,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[10626,8898,6784,3779,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11881,10282,8461,6222,2406,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13910,12478,10944,9232,7214,4246,948,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16195,14872,13503,12066,10525,8804,6750,3625,826,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18284,16987,15697,14381,13031,11617,10104,8431,6430,3408,690,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19966,18636,17336,16055,14772,13473,12141,10746,9252,7583,5564,2398,20,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21188,19764,18407,17109,15837,14575,13303,12013,10673,9259,7708,5899,3378,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21852,20364,18969,17649,16369,15095,13853,12589,11304,9963,8543,6966,5061,2054,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21832,20344,18948,17625,16346,15078,13835,12573,11290,9949,8529,6950,5055,2059,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[21197,19771,18423,17125,15851,14584,13311,12025,10681,9267,7712,5903,3374,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19965,18626,17326,16048,14771,13477,12146,10749,9250,7586,5563,2400,22,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18293,17005,15707,14395,13036,11619,10105,8428,6432,3407,684,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16184,14862,13501,12061,10522,8800,6743,3623,829,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13915,12486,10942,9238,7215,4245,948,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[11878,10277,8462,6215,2404,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[10635,8907,6782,3776,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[12104,10524,8791,6668,3637,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12973,11485,9864,8004,5664,1481,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14465,13082,11618,10025,8208,5960,2255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16235,14920,13560,12145,10627,8950,6976,4132,791,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17906,16610,15311,13993,12627,11198,9648,7912,5774,2395,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19292,17979,16689,15402,14098,12776,11398,9935,8336,6472,3883,414,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20275,18917,17606,16322,15051,13769,12464,11115,9686,8129,6335,3954,87,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20794,19389,18063,16777,15490,14233,12955,11644,10280,8830,7216,5269,2252,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20788,19387,18056,16768,15486,14229,12952,11641,10277,8824,7204,5265,2253,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[20275,18912,17603,16318,15044,13766,12458,11116,9684,8129,6337,3953,87,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19286,17974,16684,15395,14102,12775,11401,9937,8334,6471,3883,412,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17908,16618,15319,13998,12628,11191,9645,7908,5771,2397,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16232,14913,13561,12146,10629,8954,6979,4131,792,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14477,13096,11623,10024,8210,5966,2253,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12973,11492,9868,8000,5665,1485,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[12103,10521,8794,6668,3633,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[13402,11951,10370,8653,6533,3487,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13998,12597,11095,9440,7531,5074,560,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15046,13697,12277,10769,9098,7146,4518,427,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16318,15003,13660,12258,10764,9127,7229,4741,827,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17569,16281,14975,13649,12268,10801,9218,7413,5121,1327,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18613,17314,16024,14730,13419,12063,10634,9088,7334,5157,1395,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19391,18067,16780,15496,14206,12891,11535,10099,8544,6764,4489,307,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19788,18461,17165,15867,14601,13307,11982,10595,9118,7468,5480,2440,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19781,18448,17153,15861,14595,13297,11972,10583,9105,7454,5473,2445,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[19385,18066,16773,15490,14204,12896,11539,10105,8549,6771,4492,302,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18623,17323,16037,14742,13425,12066,10633,9083,7333,5155,1399,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17552,16264,14961,13635,12255,10800,9220,7412,5130,1336,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16320,15015,13670,12269,10772,9132,7234,4738,827,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15047,13696,12279,10763,9095,7140,4520,431,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13994,12590,11088,9436,7532,5076,558,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[13418,11967,10385,8666,6531,3483,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[14575,13242,11795,10203,8500,6383,3329,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14953,13606,12183,10676,8996,7043,4441,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15626,14297,12924,11472,9903,8133,5980,2398,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16433,15128,13797,12408,10937,9336,7500,5192,1020,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17267,15979,14669,13335,11940,10457,8830,6946,4464,243,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17985,16690,15396,14089,12746,11339,9835,8172,6208,3394,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18516,17218,15929,14636,13317,11961,10531,8972,7208,5016,790,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18806,17512,16205,14932,13628,12281,10875,9383,7706,5678,2629,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18792,17498,16195,14920,13621,12271,10864,9370,7692,5674,2628,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[18527,17224,15937,14647,13325,11974,10536,8978,7217,5019,793,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17982,16694,15401,14091,12749,11342,9840,8172,6205,3394,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17266,15974,14668,13334,11935,10454,8825,6947,4466,242,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16424,15124,13789,12402,10930,9327,7507,5189,1021,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15620,14295,12924,11471,9905,8136,5977,2404,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14955,13601,12182,10672,8994,7044,4442,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[14583,13253,11807,10214,8512,6380,3321,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]],

[[15711,14365,13049,11598,10011,8325,6211,3158,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15880,14571,13216,11797,10263,8570,6561,3761,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16199,14886,13538,12137,10649,9012,7121,4660,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16589,15286,13967,12591,11143,9571,7800,5627,1807,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16992,15706,14391,13043,11634,10125,8454,6490,3769,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17394,16095,14792,13465,12083,10617,9022,7192,4863,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17632,16351,15055,13734,12369,10933,9395,7652,5538,2012,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17827,16500,15228,13912,12561,11142,9620,7924,5867,2809,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17796,16488,15217,13894,12553,11127,9602,7903,5861,2807,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17661,16365,15066,13765,12380,10951,9414,7670,5541,2016,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17365,16079,14785,13437,12068,10598,9006,7182,4861,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[17012,15715,14399,13055,11642,10134,8462,6492,3771,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16582,15280,13951,12581,11131,9563,7793,5624,1806,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[16189,14886,13547,12144,10655,9020,7125,4667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15886,14573,13213,11793,10258,8563,6554,3755,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
[15713,14377,13063,11613,10025,8339,6215,3156,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]];

1;

__END__

This is the code used to generate this table:

#!/usr/bin/perl

use strict;
use warnings;
use 5.010;

use Astro::Coords;
use Astro::Telescope;
use DateTime;
use YAML;

# Creates an occlusion mask for 64 different latitudes (starting at the south pole).
# Each occlusion mask covers 16 azimuth directions (starting north).
# For each direction 32 elevations are covered from horizon to zenith.
# Values are number of minutes of possible sunlight per year in that direction given an occlusion at that elevation.

my $all;

for my $latitude_int (-32 .. 31)
{
    my $latitude = 1.5707963 * $latitude_int / 32;

    my $c = new Astro::Coords (planet => 'sun');

    # Latitude and Longitude in radians
    $c->telescope (new Astro::Telescope (Name => 'My Location', Long => 0.0, Lat => $latitude, Alt => 0.0));

    my $table = [];

    # sample every 60 seconds for a year
    for my $second (0 .. 60*60*24*365.25/60)
    {
        $second *= 60;
        my $dt = DateTime->from_epoch (epoch => $second);
        $c->datetime ($dt);

        my ($az, $el) = $c->azel;

        # sixteen lateral buckets
        my $az_int = int (16 * $az->radians / 6.2831853);

        my $el_rad = $el->radians;
        next unless $el_rad > 0;

        # thirty-two vertical buckets from horizon to zenith
        my $el_int = int (32 * $el_rad / 1.5707963);
        $table->[$az_int]->[$el_int]++;

    }

    # create the lookup table for use with an occlusion mask
    my $combination;
    for my $vert (0 .. 31)
    {
        for my $horiz (0 .. 15)
        {
            for my $elev (0 .. $vert)
            {
                $combination->[$horiz]->[$elev] += $table->[$horiz]->[$vert];
            }
        }
    }

    push @{$all}, $combination;

}

YAML::DumpFile ('sun.yml', $all);

0;
